from __future__ import annotations

from typing import Any, overload, Callable, TYPE_CHECKING

from mehtap.control_structures import LuaError
from mehtap.operations import index, call
from mehtap.values import (
    LuaNil,
    LuaBool,
    LuaNumber,
    LuaString,
    LuaTable,
    LuaFunction,
    LuaValue, type_of_lv,
)
from mehtap.vm import VirtualMachine


if TYPE_CHECKING:
    from mehtap.values import LuaNilType


@overload
def lua2py(value: LuaNilType) -> None: ...


@overload
def lua2py(value: LuaBool) -> bool: ...


@overload
def lua2py(value: LuaNumber) -> int | float: ...


@overload
def lua2py(value: LuaString) -> bytes: ...


@overload
def lua2py(value: LuaTable) -> dict: ...


@overload
def lua2py(value: LuaFunction) -> Callable: ...


@overload
def lua2py(value: LuaValue) -> Any: ...


def lua2py(value: Any) -> Any:
    """Convert a :class:`LuaValue` to a plain Python value.

    If the value has a ``__py`` metamethod,
    the converter will call it and convert its return value instead.

    Sequence tables are not converted to lists, they are also converted to
    dicts having keys *1, 2, ..., n*.

    :class:`LuaFunctions <LuaFunction>` are converted using a wrapper function,
    which converts all arguments into :class:`LuaValues <LuaValue>`,
    calls the :class:`LuaFunction` using them,
    and then converts the return value back to a Python value.

    This function is implemented using memoization,
    so it can convert recursive data structures.
    """
    return _lua2py(value, {})


PY_SYMBOL = LuaString(b"__py")


def _lua2py(lua_val, memos):
    if id(lua_val) in memos:
        return memos[id(lua_val)]
    if lua_val is LuaNil:
        return None
    if isinstance(lua_val, LuaBool):
        return lua_val.true
    if isinstance(lua_val, LuaNumber):
        return lua_val.value
    if isinstance(lua_val, LuaString):
        return lua_val.content
    if isinstance(lua_val, LuaTable):
        mt = lua_val.get_metatable()
        if mt is not LuaNil:
            metamethod = index(mt, PY_SYMBOL)
            if metamethod is not LuaNil:
                if not isinstance(metamethod, LuaFunction):
                    raise LuaError(
                        f"metavalue __py is a {type_of_lv(metamethod)}, "
                        f"not a function"
                    )
                if metamethod.parent_scope or not metamethod.gets_scope:
                    m = _lua2py(
                        call(
                            metamethod,
                            [lua_val],
                            metamethod.parent_scope,
                        ),
                        memos,
                    )
                else:
                    vm = VirtualMachine()
                    m = _lua2py(
                        call(
                            metamethod,
                            [lua_val],
                            vm.root_scope,
                        ),
                        memos,
                    )
                memos[id(lua_val)] = m
                return m
        m = {}
        memos[id(lua_val)] = m
        for k, v in lua_val.map.items():
            if v is LuaNil:
                continue
            py_v = _lua2py(v, memos)
            py_k = _lua2py(k, memos)
            m[py_k] = py_v
        return m
    if isinstance(lua_val, LuaFunction):
        from mehtap.py2lua import py2lua

        def func(*args):
            return_values = call(
                lua_val,
                args=[py2lua(x) for x in args],
                scope=None,
            )
            return [lua2py(rv) for rv in return_values]

        if lua_val.name is not None:
            func.__name__ = func.__qualname__ = lua_val.name
        else:
            func.__name__ = func.__qualname__ = "<anonymous Lua function>"
        return func
    raise TypeError(f"unknown LuaValue {lua_val!r}")



from __future__ import annotations

import subprocess
from io import SEEK_CUR, SEEK_SET, SEEK_END
from os import fsync
from tempfile import TemporaryFile
from typing import TypeVar, BinaryIO, IO

import attrs

from mehtap.ast_nodes import Numeral
from mehtap.ast_transformer import transformer
from mehtap.control_structures import LuaError
from mehtap.library.provider_abc import LibraryProvider
from mehtap.parser import numeral_parser
from mehtap.py2lua import lua_function, PyLuaRet
from mehtap.scope import Scope
from mehtap.values import (
    LuaTable,
    LuaString,
    LuaNil,
    LuaNumber,
    LuaValue,
    LuaUserdata,
    LuaIndexableABC,
    LuaFunction,
    type_of_lv,
)

FAIL = LuaNil


@attrs.define(slots=True, eq=False, repr=False)
class LuaFile(LuaUserdata, LuaIndexableABC):
    io: BinaryIO

    def rawput(self, key: LuaValue, value: LuaValue, *, raw: bool = True) -> None:
        raise LuaError("attempt to set index on a file value")

    def rawget(self, key: LuaValue, *, raw: bool = True) -> LuaValue:
        if not isinstance(key, LuaString):
            raise LuaError("invalid index to a file value")
        match key.content:
            case b"close":
                return _lf_file_method_close
            case b"flush":
                return _lf_file_method_flush
            case b"lines":
                return _lf_file_method_lines
            case b"read":
                return _lf_file_method_read
            case b"seek":
                return _lf_file_method_seek
            case b"setvbuf":
                return _lf_file_method_setvbuf
            case b"write":
                return _lf_file_method_write
            case _:
                raise LuaError("invalid index to a file value")

    T = TypeVar("T")

    def get_with_fallback(self, key: LuaValue, fallback: T) -> LuaValue | T:
        return self.rawget(key)

    def has(self, key: LuaValue) -> bool:
        return isinstance(key, LuaString) and key.content in (
            b"close",
            b"flush",
            b"lines",
            b"read",
            b"seek",
            b"setvbuf",
            b"write",
        )


@lua_function(name="close")
def _lf_file_method_close(self: LuaFile, /) -> PyLuaRet:
    self.io.close()
    return None


@lua_function(name="flush")
def _lf_file_method_flush(self: LuaFile, /) -> PyLuaRet:
    return _fsync_io(self.io)


def _fsync_io(io: IO, /) -> PyLuaRet:
    io.flush()
    fsync(io.fileno())
    return None


@lua_function(name="lines", gets_scope=True)
def _lf_file_method_lines(scope: Scope, self: LuaFile, /, *formats) -> PyLuaRet:
    return _file_method_lines(scope, self, *formats)


def _file_method_lines(scope: Scope, self: LuaFile, /, *formats) -> PyLuaRet:
    # When no format is given, uses "l" as a default.
    if not formats:
        formats = (LuaString(b"l"),)

    # Returns an iterator function that,
    @lua_function()
    def iterator_function() -> PyLuaRet:
        # each time it is called, reads the file
        # according to the given formats.
        return _file_method_read(scope, self, *formats)

    # As an example, the construction
    #     for c in file:lines(1) do body end
    # will iterate over all characters of the file, starting at the current
    # position.
    # Unlike io.lines, this function does not close the file when the loop ends.
    return [iterator_function]


def _read_format_n(scope: Scope, file: LuaFile) -> LuaNumber | FAIL:
    # "n": reads a numeral and returns it as a float or an integer,
    # following the lexical conventions of Lua.
    # (The numeral may have leading whitespaces and a sign.) This format
    # always reads the longest input sequence that is a valid prefix for a
    # numeral; if that prefix does not form a valid numeral
    # (e.g., an empty string, "0x", or "3.4e-") or it is too long
    # (more than 200 characters), it is discarded and the format returns
    # fail.
    part = file.io.read(201)
    if not part:
        return FAIL
    for length in range(len(part), 0, -1):
        try:
            input = part[:length].decode("ascii")
            parsed = numeral_parser.parse(input)
            transformed: Numeral = transformer.transform(parsed)
            return transformed.evaluate(scope=scope)
        except Exception:
            continue
        finally:
            file.io.seek(201 - len(part), SEEK_CUR)
            if length == 201:
                return FAIL
    file.io.seek(-len(part), SEEK_CUR)
    return FAIL


def _read_format_a(file: LuaFile) -> LuaString:
    # "a": reads the whole file, starting at the current position. On end of
    # file, it returns the empty string; this format never fails.
    return LuaString(file.io.read())


def _read_format_l(file: LuaFile) -> LuaString | FAIL:
    # "l": reads the next line skipping the end of line, returning fail on
    # end of file. This is the default format.
    line = file.io.readline()
    if not line:
        return FAIL
    return LuaString(line[:-1])


def _read_format_big_l(file: LuaFile) -> LuaString:
    # "L": reads the next line keeping the end-of-line character (if present),
    # returning fail on end of file.
    line = file.io.readline()
    if not line:
        return FAIL
    return LuaString(line)


def _read_format_number(file: LuaFile, number: int) -> LuaString:
    # number: reads a string with up to this number of bytes, returning fail
    # on end of file.
    # If number is zero,
    if number == 0:
        # it reads nothing and returns an
        # empty string, or fail on end of file.
        check_c = file.io.read(1)
        if not check_c:
            return FAIL
        file.io.seek(-1, SEEK_CUR)
        return LuaString(b"")
    acc = file.io.read(number)
    if not acc:
        return FAIL
    return LuaString(acc)


@lua_function(name="read", gets_scope=True)
def _lf_file_method_read(scope: Scope, self: LuaFile, /, *formats: LuaValue) \
        -> PyLuaRet:
    return _file_method_read(scope, self, *formats)


def _file_method_read(scope: Scope, self: LuaFile, /, *formats: LuaValue) \
        -> PyLuaRet:
    if not formats:
        return [_read_format_l(self)]
    return_vals: list[LuaValue] = []
    for format in formats:
        if isinstance(format, LuaNumber):
            return_vals.append(_read_format_number(self, int(format.value)))
        elif format == LuaString(b"n"):
            return_vals.append(_read_format_n(scope=scope, file=self))
        elif format == LuaString(b"a"):
            return_vals.append(_read_format_a(self))
        elif format == LuaString(b"l"):
            return_vals.append(_read_format_l(self))
        elif format == LuaString(b"L"):
            return_vals.append(_read_format_big_l(self))
        else:
            raise LuaError(
                "invalid format (valid formats are a number, 'n', 'a', 'l', "
                "and 'L'.)"
            )
    return return_vals


SYMBOL__FD = LuaString(b"__fd")


@lua_function(name="seek")
def _lf_file_method_seek(
    self: LuaFile,
    whence: LuaString | None = None,
    offset: LuaNumber | None = None,
    /,
) -> PyLuaRet:
    if whence is None:
        whence = LuaString(b"cur")
    if offset is None:
        offset = LuaNumber(0)
    match whence.content:
        case b"set":
            new_offset = self.io.seek(int(offset.value), SEEK_SET)
        case b"cur":
            new_offset = self.io.seek(int(offset.value), SEEK_CUR)
        case b"end":
            new_offset = self.io.seek(int(offset.value), SEEK_END)
        case _:
            raise LuaError("invalid whence (must be 'set', 'cur', or 'end'.)")
    return [LuaNumber(new_offset)]


@lua_function(name="setvbuf", gets_scope=True)
def _lf_file_method_setvbuf(
    scope: Scope, self: LuaFile, mode, size=None, /
) -> PyLuaRet:
    scope.vm.get_warning("file:setvbuf(): ignored call")
    return None


@lua_function(name="write")
def _lf_file_method_write(
    self: LuaFile,
    /,
    *values: LuaValue,
) -> PyLuaRet:
    return _file_method_write(self, *values)


def _file_method_write(
    self: LuaFile,
    /,
    *values: LuaValue,
) -> PyLuaRet:
    for value in values:
        if isinstance(value, LuaString):
            self.io.write(value.content)
        elif isinstance(value, LuaNumber):
            self.io.write(str(value).encode("ascii"))
    return [self]


@lua_function(name="close", gets_scope=True)
def lf_io_close(scope: Scope, file: LuaFile | None = None, /) -> PyLuaRet:
    return io_close(scope, file)


def io_close(scope: Scope, file: LuaFile | None = None, /) -> PyLuaRet:
    if file is None:
        binary_io = scope.vm.default_output
    else:
        binary_io = file.io
    binary_io.close()
    return None


@lua_function(name="flush", gets_scope=True)
def lf_io_flush(scope: Scope, /) -> PyLuaRet:
    return io_flush(scope)


def io_flush(scope: Scope, /) -> PyLuaRet:
    _fsync_io(scope.vm.default_output)
    return None


@lua_function(name="input", gets_scope=True)
def lf_io_input(
    scope: Scope, file: LuaFile | LuaString | None = None, /
) -> PyLuaRet:
    return io_input(scope, file)


def io_input(scope: Scope, file: LuaFile | LuaString | None = None, /):
    # When called with a file name, it opens the named file (in text mode),
    # and sets its handle as the default input file.
    try:
        if isinstance(file, LuaString):
            file = open(file.content, "rb")
            scope.vm.default_input = file
            return None
        # When called with a file handle, it simply sets this file handle as
        # the default input file.
        if isinstance(file, LuaFile):
            scope.vm.default_input = file.io
            return None
        # When called without arguments, it returns the
        # current default input file.
        if file is None:
            return [LuaFile(scope.vm.default_input)]
        # In case of errors this function raises the error, instead of
        # returning an error code.
    except Exception as e:
        raise LuaError(f"io.input(): {e!s}")


@lua_function(name="lines", gets_scope=True)
def lf_io_lines(
    scope: Scope, filename: LuaString | None = None, /, *formats
) -> PyLuaRet:
    return io_lines(scope, filename, *formats)


def io_lines(
    scope: Scope, filename: LuaString | None = None, /, *formats
) -> PyLuaRet:
    # The call io.lines() (with no file name) is equivalent to
    # io.input():lines("l"); that is, it iterates over the lines of the
    # default input file.
    # In this case, the iterator does not close the file when the loop ends.
    if filename is None:
        file_handle = LuaFile(scope.vm.default_input)
        to_close = False
    else:
        file_handle = LuaFile(open(filename.content, "rb"))
        to_close = True

    # Opens the given file name in read mode and returns an iterator
    # function that works like file:lines(···) over the opened file.
    # When the iterator function fails to read any value, it automatically
    # closes the file.
    @lua_function()
    def iterator_function() -> PyLuaRet:
        f = _file_method_read(scope, file_handle, *formats)
        if to_close and (
            f is None
            or any(r is LuaNil for r in f)
        ):
            file_handle.io.close()
        return f
    # Besides the iterator function, io.lines returns three other
    # values: two nil values as placeholders, plus the created file
    # handle.
    # Therefore, when used in a generic for loop, the file is closed
    # also if the loop is interrupted by an error or a break.
    return [iterator_function, LuaNil, LuaNil, file_handle]


@lua_function(name="open")
def lf_io_open(
    filename: LuaString, mode: LuaString | None = None, /
) -> PyLuaRet:
    return io_open(filename, mode)


def io_open(filename: LuaString, mode: LuaString | None = None, /) -> PyLuaRet:
    # This function opens a file, in the mode specified in the string mode.
    # In case of success, it returns a new file handle.
    #
    # The mode string can be any of the following:
    #
    #     "r": read mode (the default);
    #     "w": write mode;
    #     "a": append mode;
    #     "r+": update mode, all previous data is preserved;
    #     "w+": update mode, all previous data is erased;
    #     "a+": append update mode, previous data is preserved, writing is
    #           only allowed at the end of file.
    #
    # The mode string can also have a 'b' at the end, which is needed in
    # some systems to open the file in binary mode.
    if mode is None:
        mode = LuaString(b"r")
    if not isinstance(mode, LuaString):
        raise LuaError("'mode' must be a string")
    match mode.content:
        case b"r" | b"rb":
            mode_str = "rb"
        case b"w" | b"wb":
            mode_str = "wb"
        case b"a" | b"ab":
            mode_str = "ab"
        case b"r+" | b"r+b":
            mode_str = "r+b"
        case b"w+" | b"w+b":
            mode_str = "w+b"
        case b"a+" | b"a+b":
            mode_str = "a+b"
        case _:
            raise LuaError("invalid mode (must match /[rwa]+?b?/.)")
    return [LuaFile(open(filename.content, mode_str))]


@lua_function(name="output", gets_scope=True)
def lf_io_output(
    scope: Scope, file: LuaFile | LuaString | None = None, /
) -> PyLuaRet:
    return io_output(scope, file)


def io_output(
    scope: Scope, file: LuaFile | LuaString | None = None, /):
    # Similar to io.input, but operates over the default output file.

    # When called with a file name, it opens the named file (in text mode),
    # and sets its handle as the default input file.
    try:
        if isinstance(file, LuaString):
            scope.vm.default_output = open(file.content, "rb")
            return None
        # When called with a file handle, it simply sets this file handle as
        # the default input file.
        if isinstance(file, LuaFile):
            scope.vm.default_output = file.io
            return None
        # When called without arguments, it returns the
        # current default input file.
        if file is None:
            return [LuaFile(scope.vm.default_output)]
        # In case of errors this function raises the error, instead of
        # returning an error code.
    except Exception as e:
        raise LuaError(f"io.output(): {e!s}")


@lua_function(name="popen")
def lf_io_popen(prog, mode=None, /) -> PyLuaRet:
    return io_popen(prog, mode)


def io_popen(prog, mode=None, /):
    # io.popen (prog [, mode])
    # This function is system dependent and is not available on all
    # platforms.
    # Starts the program prog in a separated process and returns a file
    # handle that you can use to read data from this program
    # (if mode is "r", the default) or to write data to this program
    # (if mode is "w").
    if not isinstance(prog, LuaString):
        type_of_prog = type_of_lv(prog)
        raise LuaError(
            f"bad argument #1 to 'popen' "
            f"(string expected, got {type_of_prog})"
        )
    if mode is not None:
        if not isinstance(mode, LuaString):
            type_of_mode = type_of_lv(mode)
            raise LuaError(
                f"bad argument #2 to 'popen' "
                f"(string expected, got {type_of_mode})"
            )
        if mode.content not in (b"r", b"w"):
            raise LuaError(
                f"invalid argument #2 to popen ('r' or 'w' expected)"
            )
        mode_str = mode.content.decode("ascii")
    else:
        mode_str = "r"
    popen = subprocess.Popen(
        prog.content,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=False,
    )
    if mode_str == "r":
        return [LuaFile(popen.stdout)]
    return [LuaFile(popen.stdin)]


@lua_function(name="read", gets_scope=True)
def lf_io_read(scope: Scope, /, *formats) -> PyLuaRet:
    return io_read(scope, *formats)


def io_read(scope: Scope, /, *formats) -> PyLuaRet:
    # io.read (···)
    #
    # Equivalent to io.input():read(···).
    return _file_method_read(scope, io_input(scope)[0])


@lua_function(name="tmpfile")
def lf_io_tmpfile() -> PyLuaRet:
    return io_tmpfile()


def io_tmpfile() -> PyLuaRet:
    # io.tmpfile ()
    #
    # In case of success, returns a handle for a temporary file. This file
    # is opened in update mode and it is automatically removed when the
    # program ends.
    return [LuaFile(TemporaryFile())]


@lua_function(name="type")
def lf_io_type(obj: LuaValue, /) -> PyLuaRet:
    return io_type(obj)


def io_type(obj: LuaValue, /) -> PyLuaRet:
    # io.type (obj)
    #
    # Checks whether obj is a valid file handle. Returns the string "file"
    # if obj is an open file handle, "closed file" if obj is a closed file
    # handle, or fail if obj is not a file handle.
    if not isinstance(obj, LuaFile):
        return [FAIL]
    if obj.io.closed:
        return [LuaString(b"closed file")]
    return [LuaString(b"file")]


@lua_function(name="write", gets_scope=True)
def lf_io_write(scope: Scope, /, *values: LuaValue) -> PyLuaRet:
    return io_write(scope, *values)


def io_write(scope: Scope, /, *values: LuaValue) -> PyLuaRet:
    # io.write (···)
    #
    # Equivalent to io.output():write(···).
    return _file_method_write(LuaFile(scope.vm.default_output), *values)


SYMBOL_IO = LuaString(b"io")


class IOLibrary(LibraryProvider):
    def provide(self, global_table: LuaTable) -> None:
        io_table = LuaTable()
        global_table.rawput(SYMBOL_IO, io_table)

        for name_of_global, value_of_global in globals().items():
            if name_of_global.startswith("lf_io_"):
                assert isinstance(value_of_global, LuaFunction)
                assert value_of_global.name
                io_table.rawput(
                    LuaString(value_of_global.name.encode("ascii")),
                    value_of_global,
                )



from pathlib import Path

import lark


with open(Path(__file__).parent / "lua.lark", "r", encoding="utf-8") as f:
    lua_grammar = f.read()

chunk_parser = lark.Lark(
    lua_grammar,
    start="chunk",
    parser="earley",
    propagate_positions=True,
    debug=True,
)

expr_parser = lark.Lark(
    lua_grammar,
    start="exp",
    parser="earley",
    propagate_positions=True,
    debug=True,
)

numeral_parser = lark.Lark(
    lua_grammar,
    start="numeral",
    parser="earley",
    propagate_positions=True,
    debug=True,
)



from __future__ import annotations

from mehtap.library.stdlib.io_library import IOLibrary
from mehtap.library.stdlib.os_library import OSLibrary
from mehtap.library.stdlib.basic_library import BasicLibrary
from mehtap.values import LuaTable


def create_global_table() -> LuaTable:
    global_table = LuaTable()

    BasicLibrary().provide(global_table)
    OSLibrary().provide(global_table)
    IOLibrary().provide(global_table)

    return global_table



from __future__ import annotations

from collections.abc import Sequence, MutableSequence
from locale import strcoll
from typing import TypeAlias, TYPE_CHECKING

from mehtap.control_structures import LuaError
from mehtap.values import (
    LuaBool,
    LuaValue,
    LuaString,
    LuaNumber,
    MAX_INT64,
    LuaNumberType,
    MIN_INT64,
    SIGN_BIT,
    ALL_SET,
    LuaNil,
    LuaTable,
    LuaFunction,
    LuaThread,
    LuaUserdata, LuaIndexableABC, LuaCallableABC, type_of_lv,
)


if TYPE_CHECKING:
    from mehtap.scope import Scope


SYMBOL__EQ = LuaString(b"__eq")
SYMBOL__LEN = LuaString(b"__len")


def check_metamethod_binary(a: LuaValue, b: LuaValue, mm_name: LuaString) \
        -> LuaValue | None:
    mm = a.get_metavalue(mm_name)
    if mm is None:
        mm = b.get_metavalue(mm_name)
        if mm is None:
            return None
    # mm is not None
    return adjust_to_one(call(mm, args=[a, b], scope=None))


def check_metamethod_unary(a: LuaValue, mm_name: LuaString) \
        -> LuaValue | None:
    mm = a.get_metavalue(mm_name)
    if mm is None:
        return None
    return adjust_to_one(call(mm, args=[a], scope=None))


def rel_eq(a: LuaValue, b: LuaValue, *, raw: bool = False) -> LuaBool:
    """
    :param raw: Whether to bypass the ``__eq`` metamethod.
    :return: The result of ``a == b`` in Lua.
    """
    # Equality (==) first compares the type of its operands.
    # If the types are different, then the result is false.
    if type(a) is not type(b):
        return LuaBool(False)
    # Otherwise, the values of the operands are compared.
    # Strings are equal if they have the same byte content.
    if isinstance(a, LuaString):
        return LuaBool(a.content == b.content)
    # Numbers are equal if they denote the same mathematical value.
    if isinstance(a, LuaNumber):
        return LuaBool(a.value == b.value)
    # mehtap extension: All LuaBool(False) _and true_ objects are currently not
    #                   all the same object :(
    if isinstance(a, LuaBool):
        return LuaBool(a.true == b.true)
    # Tables, userdata, and threads are compared by reference:
    # two objects are considered equal only if they are the same object.
    # You can change the way that Lua compares tables and userdata by using the
    # __eq metamethod (see §2.4).
    mt_types = (LuaTable, LuaUserdata)
    if isinstance(a, mt_types) and isinstance(b, mt_types) and not raw:
        mm_res = check_metamethod_binary(a, b, SYMBOL__EQ)
        if mm_res is not None:
            return coerce_to_bool(mm_res)
    return LuaBool(a is b)


def rel_ne(a: LuaValue, b: LuaValue, *, raw: bool = False) -> LuaBool:
    """
    :param raw: Whether to bypass the ``__eq`` metamethod.
    :return: The result of ``a ~= b`` in Lua.
    """
    # The operator ~= is exactly the negation of equality (==).
    return LuaBool(not rel_eq(a, b, raw=raw).true)


SYMBOL__LT = LuaString(b"__lt")


def rel_lt(a: LuaValue, b: LuaValue) -> LuaBool:
    """
    :return: The result of ``a < b`` in Lua.
    """
    # The order operators work as follows.
    # If both arguments are numbers,
    if isinstance(a, LuaNumber) and isinstance(b, LuaNumber):
        # then they are compared according to their mathematical values,
        # regardless of their subtypes.
        return LuaBool(a.value < b.value)
    # Otherwise, if both arguments are strings,
    # then their values are compared according to the current locale.
    if isinstance(a, LuaString) and isinstance(b, LuaString):
        return LuaBool(strcoll(a.content, b.content) < 0)
    # Otherwise, Lua tries to call the __lt or the __le metamethod (see §2.4).
    mm_res = check_metamethod_binary(a, b, SYMBOL__LT)
    if mm_res is not None:
        return coerce_to_bool(mm_res)
    a_type = type_of_lv(a)
    b_type = type_of_lv(b)
    raise LuaError(f"attempt to compare {a_type} with {b_type}")


def rel_gt(a: LuaValue, b: LuaValue) -> LuaBool:
    """
    :return: The result of ``a > b`` in Lua.
    """
    # a > b is translated to b < a
    return rel_lt(b, a)


SYMBOL__LE = LuaString(b"__le")


def rel_le(a: LuaValue, b: LuaValue) -> LuaBool:
    """
    :return: The result of ``a <= b`` in Lua.
    """
    # The order operators work as follows.
    # If both arguments are numbers,
    if isinstance(a, LuaNumber) and isinstance(b, LuaNumber):
        # then they are compared according to their mathematical values,
        # regardless of their subtypes.
        return LuaBool(a.value <= b.value)
    # Otherwise, if both arguments are strings,
    # then their values are compared according to the current locale.
    if isinstance(a, LuaString) and isinstance(b, LuaString):
        return LuaBool(strcoll(a.content, b.content) <= 0)
    # Otherwise, Lua tries to call the __lt or the __le metamethod (see §2.4).
    mm_res = check_metamethod_binary(a, b, SYMBOL__LE)
    if mm_res is not None:
        return coerce_to_bool(mm_res)
    a_type = type_of_lv(a)
    b_type = type_of_lv(b)
    raise LuaError(f"attempt to compare {a_type} with {b_type}")


def rel_ge(a: LuaValue, b: LuaValue) -> LuaBool:
    """
    :return: The result of ``a >= b`` in Lua.
    """
    # a >= b is translated to b <= a
    return rel_le(b, a)


def int_wrap_overflow(value: int) -> LuaNumber:
    """Wrap around an integer value to the range of a signed 64-bit integer.

    The value is used as-is if it can already fit in a signed 64-bit integer.
    """
    if MIN_INT64 < value < MAX_INT64:
        return LuaNumber(value, LuaNumberType.INTEGER)
    whole_val, sign = divmod(value, MAX_INT64)
    if sign & 1:
        return LuaNumber(-whole_val, LuaNumberType.INTEGER)
    return LuaNumber(whole_val, LuaNumberType.INTEGER)


def coerce_float_to_int(value: LuaNumber) -> LuaNumber:
    """Coerce a number to an integer :class:`LuaNumber` if possible.

    :raises LuaError: if the conversion fails.
    """
    if not isinstance(value, LuaNumber):
        value_type = type_of_lv(value)
        raise LuaError(f"can't coerce {value_type} value to an integer")
    if value.type is LuaNumberType.INTEGER:
        return value
    # The conversion from float to integer checks whether the float has an exact
    # representation as an integer
    # (that is, the float has an integral value
    # and it is in the range of integer representation).
    v = value.value
    if v.is_integer() and MIN_INT64 <= v <= MAX_INT64:
        # If it does, that representation is the result.
        return LuaNumber(int(v), LuaNumberType.INTEGER)
    # Otherwise, the conversion fails.
    raise LuaError("number has no integer representation")


def coerce_int_to_float(value: LuaNumber) -> LuaNumber:
    """Coerce a number to a float :class:`LuaNumber`.

    This kind of conversion never fails.
    """
    if value.type is LuaNumberType.FLOAT:
        return value
    #  In a conversion from integer to float,
    #  if the integer value has an exact representation as a float,
    #  that is the result.
    #  Otherwise, the conversion gets the nearest higher or the nearest lower
    #  representable value.
    #  This kind of conversion never fails.
    return LuaNumber(float(value.value), LuaNumberType.FLOAT)


SYMBOL__ADD = LuaString(b"__add")


def arith_add(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a + b`` in Lua.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__ADD)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to add {a_type} and {b_type} values")
    # If both operands are integers,
    if a.type == LuaNumberType.INTEGER and b.type == LuaNumberType.INTEGER:
        # the operation is performed over integers and the result is an integer.
        return int_wrap_overflow(a.value + b.value)
    # Otherwise, if both operands are numbers,
    # then they are converted to floats,
    # the operation is performed following the machine's rules for
    # floating-point arithmetic (usually the IEEE 754 standard),
    # and the result is a float.
    return LuaNumber(
        coerce_int_to_float(a).value + coerce_int_to_float(b).value,
        LuaNumberType.FLOAT,
    )


def overflow_arith_add(a: LuaValue, b: LuaValue) -> tuple[bool, LuaNumber]:
    """
    :return: a tuple *(o, r)* where *o* is a boolean indicating whether the
             addition overflows and *r* is the result of ``a + b`` in Lua.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        type_a = type_of_lv(a)
        type_b = type_of_lv(b)
        raise LuaError(f"attempt to add {type_a} and {type_b} values")
    if a.type == LuaNumberType.INTEGER and b.type == LuaNumberType.INTEGER:
        summed = a.value + b.value
        wrapped = int_wrap_overflow(summed)
        return wrapped.value != summed, wrapped
    return False, LuaNumber(
        coerce_int_to_float(a).value + coerce_int_to_float(b).value,
        LuaNumberType.FLOAT,
    )


SYMBOL__SUB = LuaString(b"__sub")


def arith_sub(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a - b`` in Lua.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__SUB)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to subtract {a_type} and {b_type} values")
    if a.type == LuaNumberType.INTEGER and b.type == LuaNumberType.INTEGER:
        return int_wrap_overflow(a.value - b.value)
    return LuaNumber(
        coerce_int_to_float(a).value - coerce_int_to_float(b).value,
        LuaNumberType.FLOAT,
    )


SYMBOL__MUL = LuaString(b"__mul")


def arith_mul(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a * b`` in Lua.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__MUL)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to multiply {a_type} and {b_type} values")
    if a.type == LuaNumberType.INTEGER and b.type == LuaNumberType.INTEGER:
        return int_wrap_overflow(a.value * b.value)
    return LuaNumber(
        coerce_int_to_float(a).value * coerce_int_to_float(b).value,
        LuaNumberType.FLOAT,
    )


SYMBOL__DIV = LuaString(b"__div")


def arith_float_div(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a / b`` in Lua, which is always a float.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__DIV)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to divide {a_type} and {b_type} values")
    # Exponentiation and float division (/) always convert their operands to
    # floats and the result is always a float.
    a_float = coerce_int_to_float(a).value
    b_float = coerce_int_to_float(b).value
    if b_float == 0:
        if a_float == 0:
            return LuaNumber(float("nan"), LuaNumberType.FLOAT)
        if a_float < 0:
            return LuaNumber(float("-inf"), LuaNumberType.FLOAT)
        return LuaNumber(float("inf"), LuaNumberType.FLOAT)
    return LuaNumber(a_float / b_float, LuaNumberType.FLOAT)


SYMBOL__IDIV = LuaString(b"__idiv")


def arith_floor_div(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a // b`` in Lua.

             The result of floor division of *a* by *b* is defined as the result
             of the division of *a* by *b*
             rounded towards minus infinity.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__IDIV)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to floor divide {a_type} and {b_type} values")
    # Floor division (//) is a division that rounds the quotient towards minus
    # infinity, resulting in the floor of the division of its operands.
    if a.type == LuaNumberType.INTEGER and b.type == LuaNumberType.INTEGER:
        return int_wrap_overflow(a.value // b.value)
    return LuaNumber(
        coerce_int_to_float(a).value // coerce_int_to_float(b).value,
        LuaNumberType.INTEGER,
    )


SYMBOL__MOD = LuaString(b"__mod")


def arith_mod(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a % b`` in Lua.

             The result of modulo is defined as the remainder of a division that
             rounds the quotient towards minus infinity (floor division).
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__MOD)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to modulo {a_type} and {b_type} values")
    # Modulo is defined as the remainder of a division that rounds the quotient
    # towards minus infinity (floor division).
    if a.type == LuaNumberType.INTEGER and b.type == LuaNumberType.INTEGER:
        return int_wrap_overflow(a.value % b.value)
    return LuaNumber(
        coerce_int_to_float(a).value % coerce_int_to_float(b).value,
        LuaNumberType.INTEGER,
    )


SYMBOL__POW = LuaString(b"__pow")


def arith_exp(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a ^ b`` in Lua, which is always a float.
    """
    # Exponentiation and float division (/) always convert their operands to
    # floats and the result is always a float.
    # Exponentiation uses the ISO C function pow,
    # so that it works for non-integer exponents too.
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__POW)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to exponentiate {a_type} and {b_type} values")
    return LuaNumber(
        coerce_int_to_float(a).value ** coerce_int_to_float(b).value,
        LuaNumberType.FLOAT,
    )


SYMBOL__UNM = LuaString(b"__unm")


def arith_unary_minus(a: LuaValue) -> LuaValue:
    """
    :return: The result of ``-a`` in Lua.
    """
    if not isinstance(a, LuaNumber):
        mm_res = check_metamethod_unary(a, SYMBOL__UNM)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        raise LuaError(f"attempt to negate a {a_type} value")
    return LuaNumber(-a.value, a.type)


def _python_int_to_int64_luanumber(x: int) -> LuaNumber:
    """Convert a Python :class:`int` to a :class:`LuaNumber`.

    The input value, ``x``, is treated as an int64.
    If the value is greater than :data:`MAX_INT64`,
        * Bit value 1<<63 is interpreted as the sign bit.
        * Bit values greater than 1<<64 are ignored.
    """
    x = x & ALL_SET
    if x & SIGN_BIT:
        return LuaNumber(-x + MAX_INT64, LuaNumberType.INTEGER)
    return LuaNumber(x, LuaNumberType.INTEGER)


SYMBOL__BOR = LuaString(b"__bor")


def bitwise_or(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a | b`` in Lua.
    """
    #  All bitwise operations convert its operands to integers (see §3.4.3),
    #  operate on all bits of those integers,
    #  and result in an integer.
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__BOR)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to bitwise or {a_type} and {b_type} values")
    a = coerce_float_to_int(a)
    b = coerce_float_to_int(b)
    return _python_int_to_int64_luanumber(a.value | b.value)


SYMBOL__BXOR = LuaString(b"__bxor")


def bitwise_xor(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a ~ b`` in Lua.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__BXOR)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to bitwise xor {a_type} and {b_type} values")
    a = coerce_float_to_int(a)
    b = coerce_float_to_int(b)
    return _python_int_to_int64_luanumber(a.value ^ b.value)


SYMBOL__BAND = LuaString(b"__band")


def bitwise_and(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a & b`` in Lua.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__BAND)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to bitwise and {a_type} and {b_type} values")
    a = coerce_float_to_int(a)
    b = coerce_float_to_int(b)
    return _python_int_to_int64_luanumber(a.value & b.value)


SYMBOL__SHL = LuaString(b"__shl")


def bitwise_shift_left(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a << b`` in Lua.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__SHL)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to shift left {a_type} and {b_type} values")
    a = coerce_float_to_int(a)
    b = coerce_float_to_int(b)
    # Both right and left shifts fill the vacant bits with zeros.
    # Negative displacements shift to the other direction;
    if b.value < 0:
        return bitwise_shift_right(a, arith_unary_minus(b))
    # displacements with absolute values equal to or higher than the number of
    # bits in an integer result in zero (as all bits are shifted out).
    if b.value >= 64:
        return LuaNumber(0, LuaNumberType.INTEGER)
    return _python_int_to_int64_luanumber(a.value << b.value)


SYMBOL__SHR = LuaString(b"__shr")


def bitwise_shift_right(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a >> b`` in Lua.
    """
    if not isinstance(a, LuaNumber) or not isinstance(b, LuaNumber):
        mm_res = check_metamethod_binary(a, b, SYMBOL__SHR)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to shift right {a_type} and {b_type} values")
    a = coerce_float_to_int(a)
    b = coerce_float_to_int(b)
    if b.value < 0:
        return bitwise_shift_left(a, arith_unary_minus(b))
    if b.value >= 64:
        return LuaNumber(0, LuaNumberType.INTEGER)
    return _python_int_to_int64_luanumber(a.value >> b.value)


SYMBOL__BNOT = LuaString(b"__bnot")


def bitwise_unary_not(a: LuaValue) -> LuaValue:
    """
    :return: The result of ``~a`` in Lua.
    """
    if not isinstance(a, LuaNumber):
        mm_res = check_metamethod_unary(a, SYMBOL__BNOT)
        if mm_res is not None:
            return mm_res
        a_type = type_of_lv(a)
        raise LuaError(f"attempt to bitwise not a {a_type} value")
    a = coerce_float_to_int(a)
    return _python_int_to_int64_luanumber(~a.value)


def coerce_to_bool(a: LuaValue) -> LuaBool:
    """Coerce a value to a boolean.

    ``false`` and ``nil`` are ``false``; everything else is ``true``.
    """
    # Like the control structures (see §3.3.4),
    # all logical operators consider both false and nil as false
    # and anything else as true.
    if a is LuaNil:
        return LuaBool(False)
    if isinstance(a, LuaBool):
        return a
    return LuaBool(True)


def logical_unary_not(a: LuaValue) -> LuaBool:
    """
    :return: The result of ``not a`` in Lua.
    """
    # The negation operator not always returns false or true.
    return LuaBool(not coerce_to_bool(a).true)


def is_false_or_nil(a: LuaValue) -> bool:
    """
    :return: :data:`True` if ``a`` is ``false`` or ``nil``, :data:`False`
             otherwise.
    """
    if a is LuaNil:
        return True
    if isinstance(a, LuaBool):
        return not a.true
    return False


def str_to_lua_string(s: str) -> LuaString:
    """Convert a Python string to a Lua string.

    The Python string is encoded in ASCII.
    """
    return LuaString(s.encode("ascii"))


SYMBOL__CONCAT = LuaString(b"__concat")


def concat(a: LuaValue, b: LuaValue) -> LuaValue:
    """
    :return: The result of ``a .. b`` in Lua.
    """
    # If both operands are strings or numbers,
    types = (LuaString, LuaNumber)
    if isinstance(a, types) and isinstance(b, types):
        # then the numbers are converted to strings in a non-specified format
        # (see §3.4.3).
        if isinstance(a, LuaNumber):
            a = str_to_lua_string(str(a))
        if isinstance(b, LuaNumber):
            b = str_to_lua_string(str(b))
        return LuaString(a.content + b.content)
    # Otherwise, the __concat metamethod is called (see §2.4).
    mm_res = check_metamethod_binary(a, b, SYMBOL__CONCAT)
    if mm_res is None:
        a_type = type_of_lv(a)
        b_type = type_of_lv(b)
        raise LuaError(f"attempt to concatenate {a_type} and {b_type} values")
    return mm_res


def length(a: LuaValue, *, raw: bool = False) -> LuaValue:
    """
    :return: The result of ``#a`` in Lua.
    """
    # The length of a string is its number of bytes.
    if isinstance(a, LuaString):
        return LuaNumber(len(a.content), LuaNumberType.INTEGER)

    if a.has_metavalue(SYMBOL__LEN) and not raw:
        mm_result = check_metamethod_unary(a, SYMBOL__LEN)
        if mm_result is not None:
            return mm_result

    if isinstance(a, LuaIndexableABC):
        border = 0
        while a.has(LuaNumber(border + 1, LuaNumberType.INTEGER)):
            border += 1
            if border == MAX_INT64:
                break
        return LuaNumber(border, LuaNumberType.INTEGER)

    type_string = type_of_lv(a)
    raise LuaError(f"attempt to get length of a {type_string} value")


SYMBOL__INDEX = LuaString(b"__index")


def index(a: LuaValue, b: LuaValue) -> LuaValue:
    mv = a.get_metavalue(SYMBOL__INDEX)
    if mv is None:
        if isinstance(a, LuaIndexableABC):
            return a.rawget(b)
        raise LuaError(f"attempt to index a {type_of_lv(a)} value")
    if isinstance(a, LuaIndexableABC) and a.has(b):
        return a.rawget(b)
    allowed_mv = (LuaFunction, LuaIndexableABC)
    if not (isinstance(mv, allowed_mv) or mv.has_metavalue(SYMBOL__INDEX)):
        raise LuaError(
            f"metavalue for '__index' must be a function, table, or a value "
            f"with an '__index' metavalue"
        )
    if isinstance(mv, LuaFunction):
        return adjust_to_one(call(mv, args=[a, b], scope=None))
    return index(mv, b)


SYMBOL__NEWINDEX = LuaString(b"__newindex")


def new_index(a: LuaValue, b: LuaValue, c: LuaValue):
    mv = a.get_metavalue(SYMBOL__NEWINDEX)
    if mv is None:
        if isinstance(a, LuaIndexableABC):
            a.rawput(b, c)
            return
        raise LuaError(f"attempt to index a {type_of_lv(a)} value")
    if isinstance(a, LuaIndexableABC) and a.has(b):
        a.rawput(b, c)
        return
    allowed_mv = (LuaFunction, LuaIndexableABC)
    if not (isinstance(mv, allowed_mv) or mv.has_metavalue(SYMBOL__NEWINDEX)):
        raise LuaError(
            f"metavalue for '__newindex' must be a function, table, or a value "
            f"with a '__newindex' metavalue"
        )
    if isinstance(mv, LuaFunction):
        call(mv, args=[a, b, c], scope=None)
        return
    new_index(mv, b, c)


SYMBOL__CALL = LuaString(b"__call")


def call(
    function: LuaValue,
    args: Multires,
    scope: Scope | None,
    *,
    modify_tb: bool = True,
) -> list[LuaValue]:
    if isinstance(function, LuaCallableABC):
        return function.rawcall(args, scope=scope, modify_tb=modify_tb)
    mv = function.get_metavalue(SYMBOL__CALL)
    if mv is None:
        raise LuaError(f"attempt to call {type_of_lv(function)} value")
    if not isinstance(mv, LuaCallableABC):
        raise LuaError(f"attempt to call {type_of_lv(mv)} value")
    return mv.rawcall([function, *args], scope=scope, modify_tb=modify_tb)


Multires: TypeAlias = "Sequence[LuaValue | Multires]"
"""
A list where each element is either a :class:`LuaValue` or
:data:`Multires`.
"""


def adjust(multires: Multires, needed: int) -> Sequence[LuaValue]:
    """
    :param multires: The multires of input values.
    :param needed: The amount of values needed.
    :return: Values adjusted to the amount of values needed according to
             `the rules on adjustment of Lua`_.

    .. _the rules on adjustment of Lua:
       https://lua.org/manual/5.4/manual.html#3.4.12
    """
    # When the list of expressions ends with a multires expression,
    # all results from that expression
    # enter the list of values before the adjustment.
    multires = [x for x in multires]  # Create a shallow copy of multires.
    if multires and isinstance(multires[-1], Sequence):
        multires.extend(multires.pop())

    # The adjustment follows these rules:
    # If there are more values than needed,
    if len(multires) > needed:
        # the extra values are thrown away;
        multires = multires[:needed]
    # if there are fewer values than needed,
    if len(multires) < needed:
        # the list is extended with nil's.
        multires.extend([LuaNil] * (needed - len(multires)))

    # When a multires expression is used in a list of expressions without being
    # the last element, ..., Lua adjusts the result list of that expression
    # to one element.
    for i, value in enumerate(multires):
        if isinstance(value, Sequence):
            multires[i] = adjust(value, 1)[0]

    return multires


def adjust_flatten(multires: Multires) -> Sequence[LuaValue]:
    """
    :return: The input multires where each element is adjusted to one value
             except for the last, which is extended to the list of previous
             values.
    """
    multires = [x for x in multires]
    if multires and isinstance(multires[-1], Sequence):
        multires.extend(multires.pop())
    for i, value in enumerate(multires):
        if isinstance(value, Sequence):
            multires[i] = adjust(value, 1)[0]
    return multires


def adjust_to_one(multires_or_value: Multires | LuaValue) -> LuaValue:
    """Adjusts a multires or a single Lua value to one value.

    If the input is a multires, it adjusts the multires to one value.
    If the input is a single Lua value, it returns the value as is.

    :param multires_or_value: The multires or single Lua value to adjust.
    :return: A single Lua value.
    """
    if isinstance(multires_or_value, Sequence):
        return adjust(multires_or_value, 1)[0]
    return multires_or_value



from __future__ import annotations

from io import SEEK_SET
from os import PathLike
from os.path import basename
from typing import TypeVar, TYPE_CHECKING

import attrs
from lark.exceptions import VisitError

from mehtap.ast_transformer import transformer
from mehtap.control_structures import LuaError
from mehtap.parser import expr_parser, chunk_parser
from mehtap.values import LuaString, Variable, LuaValue

if TYPE_CHECKING:
    from mehtap.vm import VirtualMachine


AnyPath = TypeVar("AnyPath", int, str, bytes, PathLike[str], PathLike[bytes])


@attrs.define(slots=True, repr=False)
class Scope:
    vm: VirtualMachine
    parent: Scope | None
    locals: dict[LuaString, Variable] = attrs.field(factory=dict)
    varargs: list[LuaValue] | None = None
    file: str | None = None
    line: int | None = None

    def push(
        self,
        *,
        file: str | None = None,
        line: int | None = None
    ) -> Scope:
        return Scope(self.vm, self, file=file, line=line)

    def eval(self, expr: str):
        parsed_lua = expr_parser.parse(expr)
        try:
            ast = transformer.transform(parsed_lua, filename="<eval>")
        except VisitError as e:
            le = LuaError(
                LuaString(str(e.orig_exc).encode("utf-8")),
                caused_by=e,
            )
            raise le from e
        try:
            r = ast.evaluate(self)
        except Exception as e:
            le = LuaError(
                LuaString(str(e).encode("utf-8")),
                caused_by=e,
            )
            raise le from e
        if isinstance(r, LuaValue):
            return [r]
        return r

    def exec(self, chunk: str, *, filename: str | None = None) \
            -> list[LuaValue]:
        parsed_lua = chunk_parser.parse(chunk)
        try:
            ast = transformer.transform(
                parsed_lua,
                filename=filename or "<exec>"
            )
        except VisitError as e:
            le = LuaError(
                LuaString(str(e.orig_exc).encode("utf-8")),
                caused_by=e,
            )
            raise le from e
        try:
            r = ast.block.evaluate_without_inner_scope(self)
        except Exception as e:
            le = LuaError(
                LuaString(str(e).encode("utf-8")),
                caused_by=e,
            )
            raise le from e
        return r

    def exec_file(self, file_path: AnyPath) -> list[LuaValue]:
        with open(file_path, "r", encoding="utf-8") as f:
            if f.read(1) == "#":
                f.readline()
            f.seek(0, SEEK_SET)
            filename_str = basename(file_path)
            try:
                r = self.exec(f.read(), filename=filename_str)
            except Exception as e:
                raise LuaError(
                    LuaString(str(e).encode("utf-8")),
                    caused_by=e,
                )
            except LuaError as le:
                le.push_tb("main chunk", file=filename_str, line=0)
                raise le
            return r

    def __repr__(self):
        cls_name = self.__class__.__name__
        values = ",".join(f"({k})=({v})" for k, v in self.locals.items())
        if not self.varargs:
            return f"<{cls_name} locals=[{values}]>"
        else:
            varargs = ",".join(str(v) for v in self.varargs)
            return f"<{cls_name} locals=[{values}], varargs=[{varargs}]>"

    def get_varargs(self) -> list[LuaValue] | None:
        if self.varargs is None:
            raise LuaError("cannot use '...' outside a vararg function")
        return self.varargs

    def has_ls(self, key: LuaString) -> bool:
        if key in self.locals:
            return True
        if self.parent is None:
            return False
        return self.parent.has_ls(key)

    def get_ls(self, key: LuaString) -> LuaValue:
        if key in self.locals:
            return self.locals[key].value
        if self.parent is None:
            return self.vm.get_ls(key)
        return self.parent.get_ls(key)

    def put_local_ls(self, key: LuaString, variable: Variable):
        if not isinstance(variable, Variable):
            raise TypeError(f"Expected Variable, got {type(variable)}")

        if key in self.locals and self.locals[key].constant:
            raise LuaError("attempt to change constant variable")
        self.locals[key] = variable

    def put_nonlocal_ls(self, key: LuaString, value: LuaValue):
        if key in self.locals:
            if self.locals[key].constant:
                raise LuaError("attempt to change constant variable")
            self.locals[key] = Variable(value)
            return
        if self.parent is None:
            self.vm.globals.rawput(key, value)
            return
        return self.parent.put_nonlocal_ls(key, value)



from collections.abc import Sequence

import lark
from lark import Tree

import mehtap.ast_nodes as nodes
from mehtap.ast_nodes import BinaryOperator

#  The following keywords are reserved and cannot be used as names:
#      and       break     do        else      elseif    end
#      false     for       function  goto      if        in
#      local     nil       not       or        repeat    return
#      then      true      until     while
# https://www.lua.org/manual/5.4/manual.html#3.1
INVALID_NAMES = {
    "and", "break", "do", "else", "elseif", "end",
    "false", "for", "function", "goto", "if", "in",
    "local", "nil", "not", "or", "repeat", "return",
    "then", "true", "until", "while",
}


@lark.v_args(inline=True)
class LuaTransformer(lark.Transformer):
    def __default__(self, data, children, meta):
        raise NotImplementedError(
            f"Unimplemented transformer for {data=}, {children=}, {meta=}"
        )

    def __default_token__(self, token):
        return nodes.Terminal(text=token.value)

    def _call_userfunc(self, tree: lark.Tree, new_children=None):
        node: nodes.Node = super()._call_userfunc(tree, new_children)
        if not tree.meta.empty:
            if isinstance(node, Sequence):
                for child in node:
                    child.line = tree.meta.line
            else:
                node.line = tree.meta.line
        return node

    def _call_userfunc_token(self, token: lark.Token):
        node: nodes.Node = super()._call_userfunc_token(token)
        node.line = token.line
        return node

    def transform(
        self, tree, *, filename: str | None = None
    ) -> nodes.Node:
        root: nodes.Node = super().transform(tree)
        if filename is None:
            return root
        recursive_stack = [root]
        while recursive_stack:
            node = recursive_stack.pop()
            if not isinstance(node, nodes.Node):
                continue
            node.file = filename
            for slot in node.__slots__:
                new_candidate = getattr(node, slot)
                if new_candidate is None:
                    continue
                if isinstance(new_candidate, nodes.Node):
                    recursive_stack.append(new_candidate)
                elif isinstance(new_candidate, Sequence):
                    recursive_stack.extend(new_candidate)
        return root

    @staticmethod
    def NAME(token: lark.Token):
        #  The following keywords are reserved and cannot be used as names:
        #      and       break     do        else      elseif    end
        #      false     for       function  goto      if        in
        #      local     nil       not       or        repeat    return
        #      then      true      until     while
        # https://www.lua.org/manual/5.4/manual.html#3.1
        if token.value in INVALID_NAMES:
            raise ValueError(
                f"{token.value!r} is a reserved keyword and "
                f"cannot be used as a name"
            )
        return nodes.Name(name=nodes.Terminal(text=token.value))

    @staticmethod
    def var_name(name: nodes.Name) -> nodes.VarName:
        return nodes.VarName(name=name)

    @staticmethod
    def var_index(
        base: nodes.Expression,
        index: nodes.Expression | nodes.Name,
    ) -> nodes.VarIndex:
        if isinstance(index, nodes.Name):
            return nodes.VarIndex(
                base=base,
                index=nodes.ParsedLiteralLuaStringExpr(index.as_lua_string()),
            )
        return nodes.VarIndex(base=base, index=index)

    @staticmethod
    def exp_vararg() -> nodes.VarArgExpr:
        return nodes.VarArgExpr()

    @staticmethod
    def varlist(*args: nodes.Variable) -> Sequence[nodes.Variable]:
        return args

    @staticmethod
    def tableconstructor(
        fieldlist: Sequence[nodes.FieldWithKey] | None,
    ) -> nodes.TableConstructor:
        if fieldlist is None:
            return nodes.TableConstructor(fields=tuple())
        return nodes.TableConstructor(fields=fieldlist)

    @staticmethod
    def explist(*args: nodes.Expression) -> Sequence[nodes.Expression]:
        return args

    @staticmethod
    def stat_assignment(
        varlist: list[nodes.Variable], explist: list[nodes.Expression]
    ) -> nodes.Assignment:
        return nodes.Assignment(names=varlist, exprs=explist)

    @staticmethod
    def sign(token: nodes.Terminal) -> nodes.Terminal:
        return token

    @staticmethod
    def numeral_dec(
        digits: nodes.Terminal,
        fract_digits: nodes.Terminal | None,
        e_sign: nodes.Terminal | None,
        e_digits: nodes.Terminal | None,
    ) -> nodes.NumeralDec:
        return nodes.NumeralDec(
            digits=digits,
            fract_digits=fract_digits,
            e_sign=e_sign,
            e_digits=e_digits,
        )

    @staticmethod
    def numeral_hex(
        digits: nodes.Terminal,
        fract_digits: nodes.Terminal | None,
        p_sign: nodes.Terminal | None,
        p_digits: nodes.Terminal | None,
    ) -> nodes.NumeralHex:
        return nodes.NumeralHex(
            digits=digits,
            fract_digits=fract_digits,
            p_sign=p_sign,
            p_digits=p_digits,
        )

    @staticmethod
    def field_with_key(
        key: nodes.Expression | nodes.Name,
        value: nodes.Expression,
    ) -> nodes.FieldWithKey:
        return nodes.FieldWithKey(key=key, value=value)

    @staticmethod
    def field_counter_key(value: nodes.Expression) -> nodes.FieldCounterKey:
        return nodes.FieldCounterKey(value=value)

    @staticmethod
    def fieldsep():
        return lark.Discard

    @staticmethod
    def fieldlist(*children) -> Sequence[nodes.Field]:
        if children[-1] is None:
            return children[:-1]
        return children

    @staticmethod
    def block(*children) -> nodes.Block:
        *statements, return_statement = children
        return nodes.Block(
            statements=statements,
            return_statement=return_statement,
        )

    @staticmethod
    def chunk(block: nodes.Block) -> nodes.Chunk:
        return nodes.Chunk(block=block)

    @staticmethod
    def attrib(name: nodes.Name) -> nodes.Name:
        return name

    @staticmethod
    def attname(
        name: nodes.Name, attrib: nodes.Name | None
    ) -> nodes.AttributeName:
        return nodes.AttributeName(name=name, attrib=attrib)

    @staticmethod
    def attnamelist(
        *args: nodes.AttributeName,
    ) -> Sequence[nodes.AttributeName]:
        return args

    @staticmethod
    def stat_localassignment(*children) -> nodes.LocalAssignment:
        LOCAL, attnames, values = children
        return nodes.LocalAssignment(names=attnames, exprs=values)

    @staticmethod
    def cmp_op(op: nodes.Terminal) -> nodes.Terminal:
        return op

    @staticmethod
    def shift_op(op: nodes.Terminal) -> nodes.Terminal:
        return op

    @staticmethod
    def productop(op: nodes.Terminal) -> nodes.Terminal:
        return op

    @staticmethod
    def sumop(op: nodes.Terminal) -> nodes.Terminal:
        return op

    @staticmethod
    def unop(op: nodes.Terminal) -> nodes.Terminal:
        return op

    @staticmethod
    def exp_logical_or(
        left: nodes.Expression, right: nodes.Expression
    ) -> nodes.BinaryOperation:
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator.OR,
            rhs=right,
        )

    @staticmethod
    def exp_logical_and(
        left: nodes.Expression, right: nodes.Expression
    ) -> nodes.BinaryOperation:
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator.AND,
            rhs=right,
        )

    @staticmethod
    def exp_cmp(
        left: nodes.Expression,
        operation: nodes.Terminal,
        right: nodes.Expression,
    ):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator(operation.text),
            rhs=right,
        )

    @staticmethod
    def exp_bit_or(left: nodes.Expression, right: nodes.Expression):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator.BIT_OR,
            rhs=right,
        )

    @staticmethod
    def exp_bit_xor(left: nodes.Expression, right: nodes.Expression):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator.BIT_XOR,
            rhs=right,
        )

    @staticmethod
    def exp_bit_and(left: nodes.Expression, right: nodes.Expression):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator.BIT_AND,
            rhs=right,
        )

    @staticmethod
    def exp_bit_shift(
        left: nodes.Expression,
        operation: nodes.Terminal,
        right: nodes.Expression,
    ):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator(operation.text),
            rhs=right,
        )

    @staticmethod
    def exp_concat(left: nodes.Expression, right: nodes.Expression):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator.CONCAT,
            rhs=right,
        )

    @staticmethod
    def exp_sum(
        left: nodes.Expression,
        operation: nodes.Terminal,
        right: nodes.Expression,
    ):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator(operation.text),
            rhs=right,
        )

    @staticmethod
    def exp_product(
        left: nodes.Expression,
        operation: nodes.Terminal,
        right: nodes.Expression,
    ):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator(operation.text),
            rhs=right,
        )

    @staticmethod
    def exp_unop(operation: nodes.Terminal, operand: nodes.Expression):
        return nodes.UnaryOperation(
            op=nodes.UnaryOperator(operation.text),
            exp=operand,
        )

    @staticmethod
    def exp_pow(left: nodes.Expression, right: nodes.Expression):
        return nodes.BinaryOperation(
            lhs=left,
            op=BinaryOperator.EXP,
            rhs=right,
        )

    @staticmethod
    def stat_empty():
        return nodes.EmptyStatement()

    @staticmethod
    def stat_break(BREAK):
        return nodes.Break()

    @staticmethod
    def stat_goto(GOTO, name: nodes.Name):
        return nodes.Goto(name=name)

    @staticmethod
    def label(name: nodes.Name):
        return nodes.Label(name=name)

    @staticmethod
    def retstat(_, values: Sequence[nodes.Expression]) -> nodes.ReturnStatement:
        return nodes.ReturnStatement(values=values)

    def parlist(self, namelist) -> nodes.Parlist:
        return nodes.Parlist(
            names=namelist,
            vararg=False,
        )

    def parlist_vararg(self, namelist=None) -> nodes.Parlist:
        if namelist is None:
            return nodes.Parlist(
                names=tuple(),
                vararg=True,
            )
        return nodes.Parlist(
            names=namelist,
            vararg=True,
        )

    @staticmethod
    def funcbody(parlist, block, END):
        if parlist is None:
            return nodes.FuncBody(params=tuple(), body=block)
        return nodes.FuncBody(
            params=parlist.names,
            body=block,
            vararg=parlist.vararg,
        )

    @staticmethod
    def exp_functiondef(funcbody: nodes.FuncBody) -> nodes.FuncDef:
        return nodes.FuncDef(body=funcbody)

    @staticmethod
    def funcname(*names) -> nodes.FuncName:
        method = names[-1]
        if method is not None:
            return nodes.FuncName(
                names=[*names[:-1], method],
                method=True,
            )
        return nodes.FuncName(names=names[:-1], method=False)

    @staticmethod
    def stat_function(
        FUNCTION,
        funcname: nodes.FuncName,
        funcbody: nodes.FuncBody,
    ):
        return nodes.FunctionStatement(name=funcname, body=funcbody)

    @staticmethod
    def stat_localfunction(
        LOCAL, FUNCTION, name: nodes.Name, funcbody: nodes.FuncBody
    ):
        return nodes.LocalFunctionStatement(name=name, body=funcbody)

    @staticmethod
    def stat_while(
        WHILE,
        cond: nodes.Expression,
        DO,
        block: nodes.Block,
        END,
    ):
        return nodes.While(
            condition=cond,
            block=block,
        )

    @staticmethod
    def stat_for(FOR, name, start, stop, step, DO, block, END):
        return nodes.For(
            name=name,
            start=start,
            stop=stop,
            step=step,
            block=block,
        )

    @staticmethod
    def stat_forin(FOR, namelist, IN, explist, DO, block, END):
        return nodes.ForIn(
            names=namelist,
            exprs=explist,
            block=block,
        )

    @staticmethod
    def stat_do(DO, block, END) -> nodes.Do:
        return nodes.Do(block=block)

    @staticmethod
    def elseif(exp: nodes.Expression, block: nodes.Block):
        return exp, block

    @staticmethod
    def stat_if(*args):
        cond, block, *else_ifs, else_block = args
        return nodes.If(
            blocks=[(cond, block), *else_ifs],
            else_block=else_block,
        )

    @staticmethod
    def namelist(*names: nodes.Name) -> Sequence[nodes.Name]:
        return names

    @staticmethod
    def args_list(
        explist: Sequence[nodes.Expression],
    ) -> Sequence[nodes.Expression]:
        if explist is None:
            return tuple()
        return explist

    @staticmethod
    def args_value(value: nodes.Expression) -> Sequence[nodes.Expression]:
        return (value,)

    @staticmethod
    def functioncall_regular(
        name: nodes.Expression, args: Sequence[nodes.Expression]
    ) -> nodes.FuncCallRegular:
        return nodes.FuncCallRegular(name=name, args=args)

    @staticmethod
    def functioncall_method(
        base: nodes.Expression,
        method: nodes.Name,
        args: Sequence[nodes.Expression],
    ) -> nodes.FuncCallMethod:
        return nodes.FuncCallMethod(
            object=base,
            method=method,
            args=args,
        )

    @staticmethod
    def literalstring(terminal: nodes.Terminal) -> nodes.LiteralString:
        return nodes.LiteralString(text=terminal)

    @staticmethod
    def exp_false(FALSE: nodes.Terminal) -> nodes.LiteralFalse:
        return nodes.LiteralFalse()

    @staticmethod
    def exp_true(TRUE: nodes.Terminal) -> nodes.LiteralTrue:
        return nodes.LiteralTrue()

    @staticmethod
    def exp_nil(NIL: nodes.Terminal) -> nodes.LiteralNil:
        return nodes.LiteralNil()

    @staticmethod
    def exp_parenthesized(exp: nodes.Expression) -> nodes.ParenExpression:
        return nodes.ParenExpression(exp)


transformer = LuaTransformer()



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from enum import Enum
from typing import TYPE_CHECKING, TypeVar

import attrs

from mehtap.control_structures import LuaError

if TYPE_CHECKING:
    from mehtap.ast_nodes import Block
    from mehtap.scope import Scope
    from mehtap.operations import Multires


@attrs.define(slots=True, eq=True, repr=False)
class Variable:
    """Tuple of a Lua value and its properties describing a local variable."""

    value: LuaValue
    """The value of the variable."""
    constant: bool = False
    """Whether the local is a constant."""
    to_be_closed: bool = False
    """Whether the local is a to-be-closed variable."""

    def __repr__(self):
        if self.constant:
            return f"<Variable const {self.value}>"
        if self.to_be_closed:
            return f"<Variable close {self.value}>"
        return f"<Variable var {self.value}>"


@attrs.define(slots=True, eq=False)
class LuaValue(ABC):
    """Base class that all Lua values inherit from."""

    def get_metatable(self) -> LuaNilType | LuaTable:
        """
        :return: The value's metatable if it exists, or :data:`LuaNil` if not.
        """
        if hasattr(self, "_metatable") and self._metatable:
            return self._metatable
        return LuaNil

    def has_metavalue(self, name: LuaString) -> bool:
        """
        :param name: The name of the metavalue to check for.
        :return: Whether the value has a metavalue with this key.
        """
        metatable = self.get_metatable()
        if metatable is LuaNil:
            return False
        return metatable.has(name)

    def get_metavalue(self, name: LuaString) -> LuaValue | None:
        """
        :param name: The name of the metavalue to get.
        :return: The metavalue, or :data:`None` if the value doesn't have a
                 metavalue with the given name.
        """
        metatable = self.get_metatable()
        if metatable is LuaNil:
            return None
        return metatable.get_with_fallback(name, fallback=None)

    def set_metatable(self, value: LuaTable):
        """Set the value's metatable if the value can have one.

        :raises NotImplementedError: if the value can't have a metatable
        """
        if hasattr(self, "_metatable"):
            self._metatable = value
        else:
            raise LuaError(f"cannot set metatable for {type_of_lv(self)} value")

    def remove_metatable(self):
        """Removes the value's metatable if the value can have one.

        Does nothing if the value can't have a metatable
        """
        if hasattr(self, "_metatable"):
            self._metatable = None

    def __eq__(self, other) -> bool:
        """Compare this value according to Lua's rules on equality."""
        from .operations import rel_eq

        return rel_eq(self, other).true

    def __ne__(self, other) -> bool:
        from .operations import rel_ne

        return rel_ne(self, other).true

    __ne__.__doc__ = __eq__.__doc__


@attrs.define(slots=True, eq=False, repr=False)
class LuaNilType(LuaValue):
    def __str__(self) -> str:
        return "nil"

    def __repr__(self) -> str:
        return "LuaNil"

    def __hash__(self):
        return hash(None)


LuaNil = LuaNilType()
"""Value that is different from all other LuaValues.

Represents the value of the *nil* basic type in Lua.
Sole object of the ``LuaNilType`` class.
"""
del LuaNilType


@attrs.define(slots=True, eq=False)
class LuaBool(LuaValue):
    """Class representing values of the *boolean* basic type in Lua."""

    true: bool
    """Whether this value is ``true`` or ``false``."""

    def __str__(self) -> str:
        return "true" if self.true else "false"

    def __hash__(self):
        return hash(self.true)


class LuaNumberType(Enum):
    """LuaNumberType(value)
    Enumeration of the types of numbers in Lua.
    """

    INTEGER = 1
    """Number type that represents 64-bit signed integers."""
    FLOAT = 2
    """Number type that represents double-precision floating point numbers."""


MAX_INT64 = 2**63 - 1
MIN_INT64 = -(2**63)
SIGN_BIT = 1 << 63
ALL_SET = 2**64 - 1


@attrs.define(slots=True, init=False, eq=False)
class LuaNumber(LuaValue):
    """Class representing values of the *number* basic type in Lua.

    :param value: The underlying value of this number value.
    :param type: The type of this number value.
                 If provided, must match the type of ``value``.
    """

    value: int | float
    """The underlying value of this number value."""
    type: LuaNumberType | None
    """The type of this number value."""

    def __init__(
        self,
        value: int | float,
        type: LuaNumberType | None = None,
    ) -> None:
        super().__init__()
        self.value = value
        if type is None:
            if isinstance(value, int):
                self.type = LuaNumberType.INTEGER
            else:
                self.type = LuaNumberType.FLOAT
        else:
            self.type = type
            if type == LuaNumberType.INTEGER and not isinstance(value, int):
                raise ValueError("Value is not an integer")
            elif type == LuaNumberType.FLOAT and not isinstance(value, float):
                raise ValueError("Value is not a float")

    def __str__(self) -> str:
        return str(self.value)

    def __hash__(self):
        return hash(self.value)


@attrs.define(slots=True, eq=False, frozen=True)
class LuaString(LuaValue):
    """Class representing values of the *string* basic type in Lua."""

    content: bytes
    """Sequence of the bytes of the string."""

    def __str__(self) -> str:
        return self.content.decode("utf-8")

    def __hash__(self):
        return hash(self.content)


@attrs.define(slots=True, eq=False)
class LuaObject(LuaValue, ABC):
    """Base class that *Lua objects* inherit from.

    :class:`Tables <LuaTable>`,
    :class:`functions <LuaFunction>`,
    :class:`threads <LuaThread>` and
    :class:`(full) userdata values <LuaUserdata>`
    are objects.
    See `Lua 5.4 Reference Manual, Section 2.1
    <https://lua.org/manual/5.4/manual.html#2.1>`_.
    """

    def __str__(self):
        return repr(self)


class LuaCallableABC(ABC):
    """Abstract base class for objects that can be called.

    If a type that is not a subclass of this class is attempted to be called in
    execution, a :class:`LuaError` will be raised.
    """
    @abstractmethod
    def rawcall(
        self,
        args: Multires,
        scope: Scope | None,
        *,
        modify_tb: bool = True,
    ) -> list[LuaValue]:
        """Call this value.

        :param args: Arguments to pass.
        :param scope: The scope of the caller.
        :param modify_tb: Whether to add an entry to the tracebacks of errors.
        :return: the list of :class:`LuaValue` objects that the callable returns
        """
        ...


@attrs.define(slots=True, eq=False)
class LuaFunction(LuaObject, LuaCallableABC):
    """Class representing values of the *function* basic type in Lua."""

    param_names: list[LuaString]
    """The names of the parameters of the function.

    Used when binding given arguments to the block scope when calling the
    function or when pretty-displaying the function.
    """
    variadic: bool
    """Whether the function is a *variadic function*.

    That is, a function that accepts a variable number of arguments, binding
    the excess arguments to the expression "``...``".
    """
    parent_scope: Scope | None
    """The scope the function was defined in.

    The function can access any of these values as *upvalues*.
    Not applicable for functions defined from Python.
    """
    block: Block | Callable
    """The code that the function executes.

    If of type :class:`Block`, the function is implemented in Lua.
    If of type :class:`Callable`, the function is implemented in Python.

    The :meth:`call` method should be used to call the function.
    """
    gets_scope: bool = False
    """Whether the function receives the scope as its first argument.

    Only applicable for functions implemented in Python.
    """
    name: str | None = None
    """The name of the function."""
    min_req: int | None = None
    """The minimum number of required arguments.

    Only applicable for functions implemented in Python.
    Only used for pretty-displaying the function.
    """

    def _py_param_str(self, index):
        if not self.param_names or index == len(self.param_names):
            if self.variadic:
                if self.param_names:
                    return "[, ...]"
                return "[...]"
            return ""
        my_name = self.param_names[index]
        next_name = self._py_param_str(index + 1)
        if index < self.min_req:
            if index == 0:
                return f"{my_name}{next_name}"
            return f", {my_name}{next_name}"
        if index >= self.min_req:
            if index == 0:
                return f"[{my_name}{next_name}]"
            return f"[, {my_name}{next_name}]"

    def _stringify_params(self):
        if self.min_req is not None:
            return f"({self._py_param_str(0)})"
        param_names = [str(name) for name in self.param_names]
        if self.variadic:
            param_names.append("...")
        param_list = ", ".join(param_names)
        return f"({param_list})"

    def __str__(self):
        native = "native " if callable(self.block) else ""
        if not self.name:
            return (
                f"{native}function{self._stringify_params()}: {hex(id(self))}"
            )
        return (
            f"{native}function {self.name}{self._stringify_params()}: "
            f"{hex(id(self))}"
        )

    def rawcall(
        self,
        args: Multires,
        scope: Scope | None,
        *,
        modify_tb: bool = True,
    ) -> list[LuaValue]:
        """Call the function.

        :param args: The arguments to pass to the function.
        :param scope: The scope of the caller.
        :param modify_tb: Whether to add self to the traceback of exceptions.
                          If the caller is handling traceback,
                          should be set to :data:`False`,
                          otherwise should be set to :data:`True`.
        :return: A multires list of the return values of the function.

        The number of arguments will be suitably adjusted to the
        number of the function's arguments according to
        `the rules on adjustment of Lua`_.
        Can contain multires expressions.

        .. _the rules on adjustment of Lua:
           https://lua.org/manual/5.4/manual.html#3.4.12
        """
        from mehtap.control_structures import ReturnException
        from mehtap.vm import VirtualMachine
        from mehtap.control_structures import LuaError
        try:
            self._call(
                args,
                scope or self.parent_scope or VirtualMachine().root_scope,
            )
        except LuaError as le:
            if modify_tb:
                le.push_tb(str(self))
            raise le
        except ReturnException as e:
            return e.values if e.values is not None else []
        except Exception as e:
            from mehtap.control_structures import LuaError
            s_e = str(e)
            assert s_e
            le = LuaError(
                LuaString(f"{self!s}: {e!s}".encode("utf-8")),
                caused_by=e,
            )
            if modify_tb:
                le.push_tb(str(self))
            raise le from e
        return []

    def _call(
        self,
        args: Multires,
        scope: Scope,
    ):
        if not callable(self.block):
            # Function is implemented in Lua
            new_scope = self.parent_scope.push()
            param_count = len(self.param_names)

            if self.variadic:
                from mehtap.operations import adjust_flatten

                args = adjust_flatten(args)
                new_scope.varargs = args[param_count:]
                args = args[:param_count]
            from mehtap.operations import adjust

            args = adjust(args, param_count)
            for param_name, arg in zip(self.param_names, args):
                new_scope.put_local_ls(param_name, Variable(arg))
            retvals = self.block.evaluate_without_inner_scope(new_scope)
            if retvals is not None:
                from mehtap.control_structures import ReturnException

                raise ReturnException(retvals)
        else:
            # Function is implemented in Python
            from mehtap.operations import adjust_flatten

            args = adjust_flatten(args)
            try:
                if not self.gets_scope:
                    self.block(*args)
                else:
                    self.block(scope, *args)
            # Always add Python functions to tracebacks.
            # (Ignore the modify_tb parameter of cls.call().)
            except LuaError as le:
                le.push_tb(str(self), file="<Python>", line=None)
                raise le
            except Exception as e:
                le = LuaError(
                    LuaString(str(e).encode("utf-8")),
                    caused_by=e,
                )
                le.push_tb(str(self), file="<Python>", line=None)
                raise le from e


@attrs.define(slots=True, eq=False)
class LuaUserdata(LuaObject):
    """Class representing values of the *userdata* basic type in Lua.

    Users can define their own userdata types by subclassing this class.
    An example of a subclass of this class is :class:`LuaFile`.
    """
    pass


@attrs.define(slots=True, eq=False)
class LuaThread(LuaObject):
    """Class representing values of the *thread* basic type in Lua."""
    pass


class LuaIndexableABC(LuaValue, ABC):
    """Abstract base class for objects that can be indexed.

    If a type that is not a subclass of this class is attempted to be indexed in
    execution, a :class:`LuaError` will be raised.
    """
    @abstractmethod
    def rawput(self, key: LuaValue, value: LuaValue) -> None:
        """Put a key-value pair into the object."""
        ...

    @abstractmethod
    def rawget(self, key: LuaValue) -> LuaValue:
        """Get the associated value of a key from the object."""
        ...

    T = TypeVar("T")

    @abstractmethod
    def get_with_fallback(self, key: LuaValue, fallback: T) -> LuaValue | T:
        """
        Get the associated value of a key from the object, or
        :data:`fallback` if the key isn't associated with any value.
        """
        ...

    @abstractmethod
    def has(self, key: LuaValue) -> bool:
        """
        :return: :data:`True` if the object has a value associated with
                 :data:`key`, :data:`False` otherwise.
        """
        ...


@attrs.define(slots=True, eq=False, repr=False)
class LuaTable(LuaObject, LuaIndexableABC):
    """Class representing values of the *table* basic type in Lua."""

    map: dict[LuaValue, LuaValue] = attrs.field(factory=dict)
    """The key-value pairs of the table."""
    _metatable: LuaTable | None = None
    """The metatable of the table."""

    def __repr__(self):
        if not self._metatable:
            return f"<LuaTable {self!s}>"
        return f"<LuaTable {self!s} metatable={self._metatable}>"

    def __str__(self):
        return self._recursive_detecting_str(set())

    def _recursive_detecting_str(
        self,
        seen_objects: set[int],
    ) -> str:
        i = id(self)
        if i in seen_objects:
            return "{<...>}"
        seen_objects.add(i)
        pair_list = []
        for key, value in self.map.items():
            if not isinstance(key, LuaTable):
                key_str = str(key)
            else:
                key_str = key._recursive_detecting_str(seen_objects)
            if not isinstance(value, LuaTable):
                value_str = str(value)
            else:
                value_str = value._recursive_detecting_str(seen_objects)
            pair_list.append((key_str, value_str))
        return "{" + ", ".join(f"({k})=({v})" for k, v in pair_list) + "}"

    def rawput(self, key: LuaValue, value: LuaValue):
        # Warning: Do not optimize by deleting keys that are assigned LuaNil,
        # as Lua allows you to set existing fields in a table to nil while
        # traversing it by using next().

        if isinstance(key, LuaNumber):
            if key.type == LuaNumberType.FLOAT:
                if key.value == float("nan"):
                    raise LuaError("table index is NaN")
                if key.value.is_integer():
                    key = LuaNumber(int(key.value), LuaNumberType.INTEGER)

        self.map[key] = value

    def rawget(self, key: LuaValue):
        if key in self.map:
            return self.map[key]
        return LuaNil

    T = TypeVar("T")

    def get_with_fallback(self, key: LuaValue, fallback: T) -> LuaValue | T:
        f = self.map.get(key, None)
        if f is LuaNil or f is None:
            return fallback
        return f

    def has(self, key: LuaValue) -> bool:
        return key in self.map and self.map[key] is not LuaNil


def type_of_lv(a: LuaValue) -> str:
    """
    :return: The result of ``type(a)`` in Lua as a Python :class:`str`.

    This is not an operation but it is included here for ease of access.
    """
    #  Returns the type of its only argument, coded as a string.
    #  The possible results of this function are "nil"
    #  (a string, not the value nil),
    #  "number", "string", "boolean", "table", "function", "thread", and
    #  "userdata".
    if a is LuaNil:
        return "nil"
    if isinstance(a, LuaNumber):
        return "number"
    if isinstance(a, LuaString):
        return "string"
    if isinstance(a, LuaBool):
        return "boolean"
    if isinstance(a, LuaTable):
        return "table"
    if isinstance(a, LuaFunction):
        return "function"
    if isinstance(a, LuaThread):
        return "thread"
    if isinstance(a, LuaUserdata):
        return "userdata"
    raise TypeError(f"Unexpected type: {type(a)}")



from __future__ import annotations

import sys
from typing import BinaryIO

import attrs

from mehtap.global_table import create_global_table
from mehtap.scope import Scope, AnyPath
from mehtap.values import (
    LuaTable,
    LuaString,
    Variable,
    LuaValue,
)


@attrs.define(slots=True, repr=False, init=False)
class VirtualMachine:
    globals: LuaTable
    root_scope: Scope
    emitting_warnings: bool
    verbose_tb: bool
    default_input: BinaryIO
    default_output: BinaryIO

    def __init__(self):
        self.globals = create_global_table()
        self.root_scope = Scope(self, None, varargs=[])
        self.emitting_warnings = False
        if hasattr(sys.stdin, "buffer"):
            self.default_input = sys.stdin.buffer
        else:
            self.default_input = sys.stdin
        if hasattr(sys.stdout, "buffer"):
            self.default_output = sys.stdout.buffer
        else:
            self.default_output = sys.stdout
        self.verbose_tb = False

    def eval(self, expr: str):
        return self.root_scope.eval(expr)

    def exec(self, chunk: str) -> list[LuaValue]:
        return self.root_scope.exec(chunk)

    def exec_file(self, file_path: AnyPath) -> list[LuaValue]:
        return self.root_scope.exec_file(file_path)

    def has_ls(self, key: LuaString):
        return self.root_scope.has_ls(key) or self.globals.has(key)

    def get_ls(self, key: LuaString):
        if self.root_scope.has_ls(key):
            return self.root_scope.get_ls(key)
        return self.globals.rawget(key)

    def put_local_ls(self, key: LuaString, variable: Variable):
        self.root_scope.put_local_ls(key, variable)

    def put_nonlocal_ls(self, key: LuaString, variable: Variable | LuaValue):
        if isinstance(variable, Variable):
            assert not variable.constant
            assert not variable.to_be_closed
            self.globals.rawput(key, variable.value)
            return
        elif isinstance(variable, LuaValue):
            self.globals.rawput(key, variable)
            return
        else:
            raise TypeError(f"invalid key type {type(variable)}")

    def get_warning(self, *messages: str | bytes | LuaString):
        if self.emitting_warnings:
            print(f"Warning: ", *messages, sep="", file=sys.stderr)



from __future__ import annotations

import enum
import io
import string
from abc import ABC, abstractmethod
from collections.abc import Sequence, Iterable, Callable
from typing import TYPE_CHECKING, NoReturn

import attrs

import mehtap.values as m_values
from mehtap.control_structures import BreakException, GotoException, \
    ReturnException, LuaError
from mehtap.values import (
    LuaNumber,
    LuaValue,
    LuaNumberType,
    LuaString,
    MAX_INT64,
    LuaTable,
    LuaFunction, LuaIndexableABC, type_of_lv, LuaCallableABC,
)
from mehtap.operations import (
    int_wrap_overflow,
    str_to_lua_string,
    adjust,
    coerce_to_bool,
    adjust_to_one,
    adjust_flatten,
)
import mehtap.operations as m_operations

if TYPE_CHECKING:
    from mehtap.scope import Scope


def flatten(v: Iterable[LuaValue | Iterable[LuaValue]]) -> list[LuaValue]:
    result = []
    for elem in v:
        if isinstance(elem, Iterable):
            result.extend(flatten(elem))
        else:
            result.append(elem)
    return result


@attrs.define(slots=True)
class Node(ABC):
    file: str = attrs.field(kw_only=True, default="<?>")
    line: int = attrs.field(kw_only=True, default=-1)
    pass


@attrs.define(slots=True)
class Terminal(Node):
    text: str


@attrs.define(slots=True)
class NonTerminal(Node, ABC):
    pass


@attrs.define(slots=True)
class AbstractSyntaxTree:
    root: Chunk


@attrs.define(slots=True)
class Chunk(NonTerminal):
    block: Block


@attrs.define(slots=True)
class Statement(NonTerminal, ABC):
    @abstractmethod
    def _execute(self, scope: Scope) -> Sequence[LuaValue] | None:
        pass

    def execute(self, scope: Scope) -> Sequence[LuaValue] | None:
        if not scope.vm.verbose_tb:
            return self._execute(scope)
        try:
            return self._execute(scope)
        except LuaError as le:
            le.push_tb(
                f"statement {self.__class__.__name__}",
                file=self.file,
                line=self.line,
            )
            raise le
        except Exception as e:
            le = LuaError(str(e), caused_by=e)
            le.push_tb(
                f"statement {self.__class__.__name__}",
                file=self.file,
                line=self.line,
            )
            raise le


@attrs.define(slots=True)
class Expression(NonTerminal, ABC):
    @abstractmethod
    def _evaluate(self, scope: Scope) -> LuaValue | Sequence[LuaValue]:
        pass

    def evaluate(self, scope: Scope) -> LuaValue | Sequence[LuaValue]:
        if not scope.vm.verbose_tb:
            return self._evaluate(scope)
        try:
            return self._evaluate(scope)
        except LuaError as le:
            # If something is both a Statement and an Expression,
            # only add its expression part to the traceback.
            if not isinstance(self, Statement):
                le.push_tb(
                    f"expression {self.__class__.__name__}",
                    file=self.file,
                    line=self.line,
                )
            raise le
        except Exception as e:
            le = LuaError(str(e), caused_by=e)
            le.push_tb(
                f"expression {self.__class__.__name__}",
                file=self.file,
                line=self.line,
            )
            raise le

    def evaluate_single(self, scope: Scope) -> LuaValue:
        return adjust_to_one(self.evaluate(scope))


@attrs.define(slots=True)
class ParenExpression(Expression):
    exp: Expression

    def _evaluate(self, scope: Scope) -> LuaValue | Sequence[LuaValue]:
        return self.exp.evaluate_single(scope)


@attrs.define(slots=True)
class Block(Statement, Expression):
    def _evaluate(self, scope: Scope) -> list[LuaValue]:
        return self.evaluate_without_inner_scope(
            scope.push(file=self.file, line=self.line)
        )

    def _execute(self, scope: Scope) -> Sequence[LuaValue] | None:
        return self.execute_without_inner_scope(
            scope.push(file=self.file, line=self.line)
        )

    def evaluate_without_inner_scope(self, scope: Scope) -> list[LuaValue]:
        try:
            r = self.execute_without_inner_scope(scope)
        except ReturnException as re:
            if re.values:
                return re.values
            return []
        return r

    def execute_without_inner_scope(
        self, scope: Scope
    ) -> NoReturn | list[LuaValue]:
        v = None
        index = 0
        try:
            while index < len(self.statements):
                v = self.statements[index].execute(scope)
                index += 1
        except GotoException as ge:
            requested_name = ge.label.name.text
            # TODO: This sucks.
            found = False
            for index, stmt in enumerate(self.statements):
                if not isinstance(stmt, Label):
                    continue
                if stmt.name.name.text == requested_name:
                    found = True
                    break
            if not found:
                raise ge
        if self.return_statement:
            self.return_statement.execute(scope)
        if v is not None:
            return v
        return []

    statements: Sequence[Statement]
    return_statement: ReturnStatement | None = None


@attrs.define(slots=True)
class Numeral(Expression, ABC):
    pass


@attrs.define(slots=True)
class NumeralHex(Numeral):
    digits: Terminal
    fract_digits: Terminal | None = None
    p_sign: Terminal | None = None
    p_digits: Terminal | None = None

    def _evaluate(self, scope: Scope) -> LuaValue:
        if self.fract_digits or self.p_digits:
            if not self.p_sign:
                self.p_sign = Terminal("+")
            if not self.p_digits:
                self.p_digits = Terminal("0")
            if not self.fract_digits:
                self.fract_digits = Terminal("")
            whole_val = int(self.digits.text + self.fract_digits.text, 16)
            frac_val = whole_val / 16 ** len(self.fract_digits.text)
            exp_val = 2 ** int(self.p_sign.text + self.p_digits.text)
            return LuaNumber(frac_val * exp_val, LuaNumberType.FLOAT)
        # if the value overflows, it wraps around to fit into a valid integer.
        return int_wrap_overflow(int(self.digits.text, 16))


@attrs.define(slots=True)
class NumeralDec(Numeral):
    digits: Terminal
    fract_digits: Terminal | None = None
    e_sign: Terminal | None = None
    e_digits: Terminal | None = None

    def _evaluate(self, scope: Scope) -> LuaValue:
        if self.fract_digits or self.e_digits:
            if not self.e_sign:
                self.e_sign = Terminal("+")
            if not self.e_digits:
                self.e_digits = Terminal("0")
            if not self.fract_digits:
                self.fract_digits = Terminal("0")
            return LuaNumber(
                float(
                    self.digits.text
                    + "."
                    + self.fract_digits.text
                    + "e"
                    + self.e_sign.text
                    + self.e_digits.text
                ),
                LuaNumberType.FLOAT,
            )
        whole_val = int(self.digits.text)
        if whole_val > MAX_INT64:
            return LuaNumber(float(self.digits.text), LuaNumberType.FLOAT)
        return LuaNumber(whole_val, LuaNumberType.INTEGER)


@attrs.define(slots=True)
class LiteralString(Expression):
    text: Terminal

    def _simple_string(self) -> LuaString:
        bytes_io = io.BytesIO()
        currently_skipping_whitespace = False
        currently_reading_decimal = False
        currently_read_decimal: str = ""
        str_iter = iter(self.text.text[1:-1])
        for character in str_iter:
            if currently_skipping_whitespace:
                if character in string.whitespace:
                    continue
                currently_skipping_whitespace = False
            if currently_reading_decimal:
                if (
                    character in string.digits
                    and len(currently_read_decimal) < 3
                ):
                    currently_read_decimal += character
                    continue
                bytes_io.write(bytes([int(currently_read_decimal)]))
                currently_reading_decimal = False
                currently_read_decimal = ""
            if character == "\\":
                try:
                    escape_char = next(str_iter)
                except StopIteration:
                    bytes_io.write(b"\\")
                    break
                if escape_char == "a":
                    bytes_io.write(b"\a")
                elif escape_char == "b":
                    bytes_io.write(b"\b")
                elif escape_char == "f":
                    bytes_io.write(b"\f")
                elif escape_char == "n":
                    bytes_io.write(b"\n")
                elif escape_char == "r":
                    bytes_io.write(b"\r")
                elif escape_char == "t":
                    bytes_io.write(b"\t")
                elif escape_char == "v":
                    bytes_io.write(b"\v")
                elif escape_char == "\\":
                    bytes_io.write(b"\\")
                elif escape_char == '"':
                    bytes_io.write(b'"')
                elif escape_char == "'":
                    bytes_io.write(b"'")
                elif escape_char == "\n":
                    bytes_io.write(b"\n")
                elif escape_char == "z":
                    currently_skipping_whitespace = True
                    continue
                elif escape_char == "x":
                    try:
                        hex_digit_1 = next(str_iter)
                        hex_digit_2 = next(str_iter)
                    except StopIteration:
                        raise NotImplementedError()
                    bytes_io.write(bytes.fromhex(hex_digit_1 + hex_digit_2))
                elif escape_char in string.digits:
                    currently_reading_decimal = True
                    currently_read_decimal = escape_char
                    continue
                elif escape_char == "u":
                    left_brace = next(str_iter)
                    if left_brace != "{":
                        raise NotImplementedError()
                    hex_digits = []
                    while True:
                        hex_digit = next(str_iter)
                        if hex_digit == "}":
                            break
                        hex_digits.append(hex_digit)
                    hex_str = "".join(hex_digit)
                    bytes_io.write(chr(int(hex_str, 16)).encode("utf-8"))
                else:
                    raise NotImplementedError()
                continue
            bytes_io.write(character.encode("utf-8"))
        bytes_io.seek(0)
        return LuaString(bytes_io.read())

    def _long_bracket(self) -> LuaString:
        # Literals in this bracketed form can run for several lines,
        # do not interpret any escape sequences,
        # and ignore long brackets of any other level.
        level = 1
        while self.text.text[level] == "=":
            level += 1
        level -= 1
        # [====[...
        #   0123456
        # ...]====]
        # (-)654321
        symbol_len = level + 2
        bytes_io = io.BytesIO()
        str_iter = iter(self.text.text[symbol_len:-symbol_len])
        converting_end_of_line = False
        first_character_seen = False
        for character in str_iter:
            # Any kind of end-of-line sequence (carriage return, newline,
            # carriage return followed by newline, or newline followed by
            # carriage return) is converted to a simple newline.
            if character in ("\r", "\n"):
                if not first_character_seen:
                    # When the opening long bracket is immediately followed by a
                    # newline, the newline is not included in the string.
                    converting_end_of_line = True
                    first_character_seen = True
                    continue
                if converting_end_of_line:
                    continue
                first_character_seen = True
                converting_end_of_line = True
                bytes_io.write(b"\n")
                continue
            else:
                first_character_seen = True
                converting_end_of_line = False
                bytes_io.write(character.encode("utf-8"))
        bytes_io.seek(0)
        return LuaString(bytes_io.read())

    def _evaluate(self, scope: Scope) -> LuaString:
        if self.text.text[0] != "[":
            return self._simple_string()
        return self._long_bracket()


@attrs.define(slots=True)
class LiteralFalse(Expression):
    def _evaluate(self, scope: Scope) -> LuaValue:
        return m_values.LuaBool(False)


@attrs.define(slots=True)
class LiteralTrue(Expression):
    def _evaluate(self, scope: Scope) -> LuaValue:
        return m_values.LuaBool(True)


@attrs.define(slots=True)
class LiteralNil(Expression):
    def _evaluate(self, scope: Scope) -> LuaValue:
        return m_values.LuaNil


@attrs.define(slots=True)
class Name(NonTerminal):
    name: Terminal

    def as_lua_string(self):
        return str_to_lua_string(self.name.text)


@attrs.define(slots=True)
class VarArgExpr(Expression):
    def _evaluate(self, scope: Scope) -> LuaValue | Sequence[LuaValue]:
        v = scope.get_varargs()
        if v is not None:
            return v
        return []


@attrs.define(slots=True)
class Variable(Expression, ABC):
    pass


@attrs.define(slots=True)
class VarName(Variable):
    def _evaluate(self, scope: Scope) -> LuaValue:
        return scope.get_ls(str_to_lua_string(self.name.name.text))

    name: Name


@attrs.define(slots=True)
class VarIndex(Variable):
    def _evaluate(self, scope: Scope) -> LuaValue:
        return m_operations.index(
            a=self.base.evaluate_single(scope),
            b=self.index.evaluate_single(scope)
        )

    base: Expression
    index: Expression


@attrs.define(slots=True)
class TableConstructor(Expression):
    fields: Sequence[Field]

    def _evaluate(self, scope: Scope) -> LuaValue:
        table = LuaTable()
        if not self.fields:
            return table
        counter = 1
        # If the last field in the list has the form exp and the expression is a
        # multires expression,
        # then all values returned by this expression enter the list
        # consecutively (see §3.4.12).
        last_field = self.fields[-1]
        if last_field and isinstance(last_field, FieldCounterKey):
            field_iter = iter(self.fields[:-1])
        else:
            field_iter = iter(self.fields)
        for field in field_iter:
            key: LuaValue
            if isinstance(field, FieldWithKey):
                if isinstance(field.key, Name):
                    key = str_to_lua_string(field.key.name.text)
                elif isinstance(field.key, Expression):
                    key = field.key.evaluate_single(scope)
                else:
                    raise ValueError(f"{type(field.key)=}")
                table.rawput(key, field.value.evaluate_single(scope))
            elif isinstance(field, FieldCounterKey):
                key = LuaNumber(counter, LuaNumberType.INTEGER)
                counter += 1
                table.rawput(key, field.value.evaluate_single(scope))
            else:
                raise ValueError(f"{type(field)=}")
        if last_field and isinstance(last_field, FieldCounterKey):
            last_field_value = last_field.value.evaluate(scope)
            if isinstance(last_field_value, Sequence):
                for counter, val in enumerate(last_field_value, start=counter):
                    table.rawput(
                        LuaNumber(counter, LuaNumberType.INTEGER),
                        val,
                    )
            else:
                table.rawput(
                    LuaNumber(counter, LuaNumberType.INTEGER),
                    last_field_value,
                )
        return table


@attrs.define(slots=True)
class Field(NonTerminal, ABC):
    value: Expression


@attrs.define(slots=True)
class FieldWithKey(Field):
    key: Expression | Name | None = None


@attrs.define(slots=True)
class FieldCounterKey(Field):
    pass


@attrs.define(slots=True)
class Parlist(NonTerminal):
    names: Sequence[Name]
    vararg: bool = False


@attrs.define(slots=True)
class FuncBody(Expression):
    params: Sequence[Name]
    body: Block
    vararg: bool = False

    def _evaluate(self, scope: Scope) -> LuaFunction:
        return LuaFunction(
            param_names=[p.as_lua_string() for p in self.params],
            variadic=self.vararg,
            block=self.body,
            parent_scope=scope,
            gets_scope=False,
            min_req=0,  # The parameters get adjusted.
        )


@attrs.define(slots=True)
class FuncDef(Expression):
    body: FuncBody

    def _evaluate(self, scope: Scope) -> LuaValue:
        return self.body.evaluate(scope)


@attrs.define(slots=True)
class FuncCallRegular(Expression, Statement):
    def _evaluate(self, scope: Scope) -> list[LuaValue]:
        function = self.name.evaluate_single(scope)
        args = [arg.evaluate(scope) for arg in self.args]
        try:
            r = m_operations.call(function, args, scope, modify_tb=False)
        except LuaError as le:
            le.push_tb(
                f"call of {function}",
                file=self.file,
                line=self.line,
            )
            raise le
        return r

    def _execute(self, scope: Scope) -> None | list[LuaValue]:
        r = self.evaluate(scope)
        if isinstance(r, Sequence):
            if r:
                return r
            return []
        return [r]

    name: Expression
    args: Sequence[Expression]


@attrs.define(slots=True)
class FuncCallMethod(Expression, Statement):
    def _evaluate(self, scope: Scope) -> Sequence[LuaValue]:
        # A call v:name(args) is syntactic sugar for v.name(v,args),
        # except that v is evaluated only once.
        v = self.object.evaluate_single(scope)
        function = m_operations.index(
            a=v,
            b=str_to_lua_string(self.method.name.text)
        )
        args = [v, *(arg.evaluate(scope) for arg in self.args)]
        try:
            r = m_operations.call(function, args, scope, modify_tb=False)
        except LuaError as le:
            le.push_tb(
                f"method call of {function}",
                file=self.file,
                line=self.line,
            )
            raise le
        return r

    def _execute(self, scope: Scope) -> None | list[LuaValue]:
        r = self.evaluate(scope)
        if isinstance(r, Sequence):
            if r:
                return r
            return []
        return [r]

    object: Expression
    method: Name
    args: Sequence[Expression]


class UnaryOperator(enum.Enum):
    NEG = "-"
    NOT = "not"
    LENGTH = "#"
    BIT_NOT = "~"


unary_operator_functions: dict[
    UnaryOperator, Callable[[LuaValue], LuaValue]
] = {
    UnaryOperator.NEG: m_operations.arith_unary_minus,
    UnaryOperator.NOT: m_operations.logical_unary_not,
    UnaryOperator.LENGTH: m_operations.length,
    UnaryOperator.BIT_NOT: m_operations.bitwise_unary_not,
}


@attrs.define(slots=True)
class UnaryOperation(Expression):
    op: UnaryOperator
    exp: Expression

    def _evaluate(self, scope: Scope) -> LuaValue:
        return unary_operator_functions[self.op](
            self.exp.evaluate_single(scope)
        )


class BinaryOperator(enum.Enum):
    OR = "or"
    AND = "and"
    LT = "<"
    LE = "<="
    GT = ">"
    GE = ">="
    EQ = "=="
    NE = "~="
    BIT_OR = "|"
    BIT_XOR = "~"
    BIT_AND = "&"
    SHIFT_LEFT = "<<"
    SHIFT_RIGHT = ">>"
    CONCAT = ".."
    ADD = "+"
    SUBTRACT = "-"
    MULTIPLY = "*"
    FLOAT_DIV = "/"
    FLOOR_DIV = "//"
    MODULO = "%"
    EXP = "^"


binary_operator_functions: dict[
    BinaryOperator, Callable[[LuaValue, LuaValue], LuaValue]
] = {
    BinaryOperator.LT: m_operations.rel_lt,
    BinaryOperator.LE: m_operations.rel_le,
    BinaryOperator.GT: m_operations.rel_gt,
    BinaryOperator.GE: m_operations.rel_ge,
    BinaryOperator.EQ: m_operations.rel_eq,
    BinaryOperator.NE: m_operations.rel_ne,
    BinaryOperator.BIT_OR: m_operations.bitwise_or,
    BinaryOperator.BIT_XOR: m_operations.bitwise_xor,
    BinaryOperator.BIT_AND: m_operations.bitwise_and,
    BinaryOperator.SHIFT_LEFT: m_operations.bitwise_shift_left,
    BinaryOperator.SHIFT_RIGHT: m_operations.bitwise_shift_right,
    BinaryOperator.CONCAT: m_operations.concat,
    BinaryOperator.ADD: m_operations.arith_add,
    BinaryOperator.SUBTRACT: m_operations.arith_sub,
    BinaryOperator.MULTIPLY: m_operations.arith_mul,
    BinaryOperator.FLOAT_DIV: m_operations.arith_float_div,
    BinaryOperator.FLOOR_DIV: m_operations.arith_floor_div,
    BinaryOperator.MODULO: m_operations.arith_mod,
    BinaryOperator.EXP: m_operations.arith_exp,
}


@attrs.define(slots=True)
class BinaryOperation(Expression, ABC):
    lhs: Expression
    op: BinaryOperator
    rhs: Expression

    def _evaluate(self, scope: Scope) -> LuaValue:
        op = self.op
        # Both and and or use short-circuit evaluation;
        # that is, the second operand is evaluated only if necessary.
        if op == BinaryOperator.AND:
            # The conjunction operator "and" returns its first argument if this
            # value is false or nil;
            l_val = self.lhs.evaluate_single(scope)
            if l_val is m_values.LuaNil or l_val == m_values.LuaBool(False):
                return l_val
            # otherwise, and returns its second argument.
            return self.rhs.evaluate_single(scope)
        if op == BinaryOperator.OR:
            # The disjunction operator "or" returns its first argument if this
            # value is different from nil and false;
            l_val = self.lhs.evaluate_single(scope)
            if l_val is not m_values.LuaNil and l_val != m_values.LuaBool(
                False
            ):
                return l_val
            # otherwise, or returns its second argument.
            return self.rhs.evaluate_single(scope)

        left = self.lhs.evaluate_single(scope)
        right = self.rhs.evaluate_single(scope)
        return binary_operator_functions[op](left, right)


@attrs.define(slots=True)
class ReturnStatement(Statement):
    def _execute(self, scope: Scope) -> NoReturn:
        raise ReturnException(
            flatten(
                expr.evaluate(scope)
                for expr in self.values
            )
        )

    values: Sequence[Expression]


@attrs.define(slots=True)
class EmptyStatement(Statement):
    def _execute(self, scope: Scope) -> None:
        pass


@attrs.define(slots=True)
class Assignment(Statement):
    def _execute(self, scope: Scope) -> list[LuaValue]:
        values = adjust(
            [expr.evaluate(scope) for expr in self.exprs], len(self.names)
        )
        for variable, value in zip(self.names, values):
            if isinstance(variable, VarName):
                var_name = str_to_lua_string(variable.name.name.text)
                scope.put_nonlocal_ls(var_name, value)
            elif isinstance(variable, VarIndex):
                table = variable.base.evaluate_single(scope)
                if not isinstance(table, LuaIndexableABC):
                    raise LuaError(
                        f"attempt to index {type_of_lv(table)} value"
                    )
                m_operations.new_index(
                    a=table,
                    b=variable.index.evaluate_single(scope),
                    c=value,
                )
            else:
                raise ValueError(f"{type(variable)=}")
        return values

    names: Sequence[Variable]
    exprs: Sequence[Expression]


@attrs.define(slots=True)
class Label(Statement):
    def _execute(self, scope: Scope) -> None:
        pass

    name: Name


@attrs.define(slots=True)
class Break(Statement):
    def _execute(self, scope: Scope) -> None:
        raise BreakException()


@attrs.define(slots=True)
class Goto(Statement):
    def _execute(self, scope: Scope) -> None:
        raise GotoException(self.name)

    name: Name


@attrs.define(slots=True)
class Do(Statement):
    def _execute(self, scope: Scope) -> None:
        self.block.execute(scope)

    block: Block


@attrs.define(slots=True)
class While(Statement):
    def _execute(self, scope: Scope) -> None:
        new_vm = scope.push(file=self.file, line=self.line)
        try:
            while coerce_to_bool(self.condition.evaluate_single(scope)).true:
                self.block.execute_without_inner_scope(new_vm)
        except BreakException:
            pass

    condition: Expression
    block: Block


@attrs.define(slots=True)
class Repeat(Statement):
    def _execute(self, scope: Scope) -> None:
        new_vm = scope.push(file=self.file, line=self.line)
        try:
            while True:
                self.block.execute_without_inner_scope(new_vm)
                if coerce_to_bool(self.condition.evaluate_single(new_vm)).true:
                    break
        except BreakException:
            pass

    block: Block
    condition: Expression


@attrs.define(slots=True)
class If(Statement):
    def _execute(self, scope: Scope) -> None:
        for cnd, blk in self.blocks:
            if coerce_to_bool(cnd.evaluate_single(scope)).true:
                blk.execute(scope)
                return
        if self.else_block:
            self.else_block.execute(scope)

    blocks: Sequence[tuple[Expression, Block]]
    else_block: Block | None = None


@attrs.define(slots=True)
class For(Statement):
    name: Name
    start: Expression
    stop: Expression
    step: Expression | None
    block: Block

    def _execute(self, scope: Scope) -> None:
        try:
            self._execute_internal(scope)
        except BreakException:
            pass

    def _execute_internal(self, scope: Scope) -> None:
        # This for loop is the "numerical" for loop explained in 3.3.5.
        # The given identifier (Name) defines the control variable,
        # which is a new variable local to the loop body (block).
        control_varname = self.name.as_lua_string()
        # The loop starts by evaluating once the three control expressions.
        # Their values are called respectively
        # the initial value,
        initial_value = self.start.evaluate_single(scope)
        if not isinstance(initial_value, LuaNumber):
            raise LuaError("the initial value must be a number")
        # the limit,
        limit = self.stop.evaluate_single(scope)
        if not isinstance(limit, LuaNumber):
            raise LuaError("the limit value must be a number")
        # and the step. If the step is absent, it defaults to 1.
        if self.step:
            step = self.step.evaluate_single(scope)
            if not isinstance(step, LuaNumber):
                raise LuaError("the step value must be a number")
        else:
            step = LuaNumber(1, LuaNumberType.INTEGER)
        # If both the initial value and the step are integers,
        # the loop is done with integers;
        # note that the limit may not be an integer.
        is_integer_loop = (
            initial_value.type == LuaNumberType.INTEGER
            and step.type == LuaNumberType.INTEGER
        )
        if not is_integer_loop:
            # Otherwise, the three values are converted to floats
            # and the loop is done with floats.
            initial_value = m_operations.coerce_int_to_float(initial_value)
            limit = m_operations.coerce_int_to_float(limit)
            step = m_operations.coerce_int_to_float(step)
        # After that initialization, the loop body is repeated with the value of
        # the control variable going through an arithmetic progression,
        # starting at the initial value,
        # with a common difference given by the step.
        # A negative step makes a decreasing sequence;
        # a step equal to zero raises an error.
        if step.value == 0:
            raise LuaError("step must not be zero")
        # The loop continues while the value is less than or equal to the limit
        # (greater than or equal to for a negative step).
        # If the initial value is already greater than the limit
        # (or less than, if the step is negative),
        # the body is not executed.
        is_step_negative = step.value < 0
        condition_func = (
            m_operations.rel_ge if is_step_negative else m_operations.rel_le
        )
        # For integer loops, the control variable never wraps around; instead,
        # the loop ends in case of an overflow.
        # You should not change the value of the control variable during the
        # loop.
        # If you need its value after the loop, assign it to another variable
        # before exiting the loop.
        control_val = initial_value
        inner_scope = scope.push(file=self.file, line=self.line)
        while condition_func(control_val, limit).true:
            inner_scope.put_local_ls(
                control_varname, m_values.Variable(control_val)
            )
            self.block.execute_without_inner_scope(inner_scope)
            overflow, control_val = m_operations.overflow_arith_add(
                control_val, step
            )
            if overflow and is_integer_loop:
                break


@attrs.define(slots=True)
class ForIn(Statement):
    names: Sequence[Name]
    exprs: Sequence[Expression]
    block: Block

    def _execute(self, scope: Scope) -> None:
        try:
            self._execute_internal(scope)
        except BreakException:
            pass

    def _execute_internal(self, outer_scope: Scope) -> None:
        # The generic for statement works over functions, called iterators.
        # On each iteration, the iterator function is called to produce a new
        # value, stopping when this new value is nil.

        #  A for statement like
        #      for var_1, ···, var_n in explist do body end
        # works as follows.
        # The names var_i declare loop variables local to the loop body.
        body_scope = outer_scope.push(file=self.file, line=self.line)
        name_count = len(self.names)
        names = [name.as_lua_string() for name in self.names]
        for name in names:
            body_scope.put_local_ls(name, m_values.Variable(m_values.LuaNil))
        # The first of these variables is the control variable.
        control_variable_name = names[0]
        # The loop starts by evaluating explist to produce four values:
        exp_vals = adjust([exp.evaluate(outer_scope) for exp in self.exprs], 4)
        # an iterator function,
        iterator_function = exp_vals[0]
        # a state,
        state = exp_vals[1]
        # an initial value for the control variable,
        initial_value = exp_vals[2]
        body_scope.put_local_ls(
            control_variable_name, m_values.Variable(initial_value)
        )
        # and a closing value.
        closing_value = exp_vals[3]

        nil = m_values.LuaNil
        while True:
            # Then, at each iteration, Lua calls the iterator function with two
            # arguments: the state and the control variable.
            results = adjust(
                m_operations.call(
                    iterator_function,
                    [state, body_scope.get_ls(control_variable_name)],
                    outer_scope,
                ),
                name_count,
            )
            # The results from this call are then assigned to the loop
            # variables, following the rules of multiple assignments.
            for name, value in zip(names, results):
                body_scope.put_local_ls(name, m_values.Variable(value))
            # If the control variable becomes nil, the loop terminates.
            if results[0] is nil:
                break
            # Otherwise, the body is executed and the loop goes to the next
            # iteration.
            self.block.execute_without_inner_scope(body_scope)
            continue
        if closing_value is not nil:
            # The closing value behaves like a to-be-closed variable,
            # which can be used to release resources when the loop ends.
            # Otherwise, it does not interfere with the loop.
            raise NotImplementedError()


@attrs.define(slots=True)
class FuncName(NonTerminal):
    names: Sequence[Name]
    method: bool


@attrs.define(slots=True)
class FunctionStatement(Statement):
    name: FuncName
    body: FuncBody

    def _execute(self, scope: Scope) -> list[LuaFunction]:
        function = self.body.evaluate_single(scope)
        if self.name.method:
            function.param_names.insert(0, LuaString(b"self"))
        if len(self.name.names) == 1:
            name = self.name.names[0].as_lua_string()
            function.name = name
            scope.put_nonlocal_ls(name, function)
        else:
            table = scope.get_ls(self.name.names[0].as_lua_string())
            for name in self.name.names[1:-1]:
                table = m_operations.index(
                    a=table,
                    b=name.as_lua_string()
                )
            function.name = self.name.names[-1].as_lua_string()
            m_operations.new_index(
                a=table,
                b=function.name,
                c=function,
            )
        return [function]


@attrs.define(slots=True)
class LocalFunctionStatement(Statement):
    name: Name
    body: FuncBody

    def _execute(self, scope: Scope) -> list[LuaFunction]:
        # The statement
        #      local function f () body end
        # translates to
        #      local f; f = function () body end
        # not to
        #      local f = function () body end
        # (This only makes a difference
        # when the body of the function contains references to f.)
        name = self.name.as_lua_string()
        scope.put_local_ls(name, m_values.Variable(m_values.LuaNil))
        function = self.body.evaluate_single(scope)
        function.name = name
        scope.put_local_ls(name, m_values.Variable(function))
        return [function]


@attrs.define(slots=True)
class AttributeName(NonTerminal):
    name: Name
    attrib: Name | None = None


@attrs.define(slots=True)
class LocalAssignment(Statement):
    names: Sequence[AttributeName]
    exprs: Sequence[Expression]

    def _execute(self, scope: Scope) -> list[LuaValue]:
        if self.exprs:
            exp_vals = adjust(
                [exp.evaluate(scope) for exp in self.exprs],
                len(self.names)
            )
        else:
            exp_vals = [m_values.LuaNil] * len(self.names)
        used_closed = False
        for attname, exp_val in zip(self.names, exp_vals):
            var_name = attname.name.as_lua_string()
            if attname.attrib is None:
                scope.put_local_ls(var_name, m_values.Variable(exp_val))
            else:
                attrib = attname.attrib.as_lua_string()
                if attrib.content == b"close":
                    if used_closed:
                        raise NotImplementedError()
                    used_closed = True
                    scope.put_local_ls(
                        var_name, m_values.Variable(exp_val, to_be_closed=True)
                    )
                elif attrib.content == b"const":
                    scope.put_local_ls(
                        var_name, m_values.Variable(exp_val, constant=True)
                    )
                else:
                    raise LuaError(
                        f"unknown attribute '{attrib.content.decode('ascii')}'"
                    )
        return exp_vals


@attrs.define(slots=True)
class ParsedLiteralLuaStringExpr(Expression):
    """
    Not included in the Lua grammar.
    Used when a part of an expression which results in a LuaString is parsed.
    For example, the "b" in ``a.b = 1``.
    (Which is transformed to be equivalent to ``a["b"] = 1``.)
    """

    value: LuaString

    def _evaluate(self, scope: Scope) -> LuaValue | Sequence[LuaValue]:
        return self.value



from __future__ import annotations

import locale as lc
import os
import subprocess
import sys
import tempfile
import time as py_time
import datetime

from mehtap.control_structures import LuaError
from mehtap.operations import str_to_lua_string
from mehtap.py2lua import lua_function, PyLuaWrapRet, py2lua, PyLuaRet
from mehtap.library.provider_abc import LibraryProvider
from mehtap.values import (
    LuaTable,
    LuaString,
    LuaNil,
    LuaNumber,
    LuaBool,
    LuaValue,
    LuaFunction,
)

FAIL = LuaNil


def _py_wday_to_lua_wday(x: int) -> int:
    if x == 7:
        return 1
    return x + 1


def _get_day_number_of_year(date: datetime.date) -> int:
    return date.timetuple().tm_yday


def _oserror_to_errtuple(e: OSError) -> list[LuaValue]:
    return [
        LuaNil,
        LuaString(e.strerror.encode("utf-8")),
        LuaNumber(e.errno),
    ]


_str_to_lc_category_map = {
    "all": lc.LC_ALL,
    "collate": lc.LC_COLLATE,
    "ctype": lc.LC_CTYPE,
    "monetary": lc.LC_MONETARY,
    "numeric": lc.LC_NUMERIC,
    "time": lc.LC_TIME,
}


def _get_category_from_luastr(luastr: LuaString) -> int:
    string = luastr.content.decode("utf-8")
    return _str_to_lc_category_map[string]


@lua_function(name="clock")
def lf_os_clock() -> PyLuaRet:
    return os_clock()


def os_clock() -> PyLuaRet:
    # Returns an approximation of the amount in seconds of CPU time used by
    # the program, as returned by the underlying ISO C function clock.
    return [LuaNumber(py_time.process_time())]


@lua_function(name="date")
def lf_os_date(format=None, time=None, /) -> PyLuaRet:
    return os_date(format, time)


def os_date(format=None, time=None, /) -> PyLuaRet:
    raise NotImplementedError()


@lua_function(name="difftime")
def lf_os_difftime(t2, t1, /) -> PyLuaRet:
    return os_difftime(t2, t1)


def os_difftime(t2, t1, /) -> PyLuaRet:
    raise NotImplementedError()


@lua_function(name="execute")
def lf_os_execute(command=None, /) -> PyLuaRet:
    return os_execute(command)


def os_execute(command=None, /) -> PyLuaRet:
    # When called without a command, os.execute returns a boolean that is
    # true if a shell is available.
    if command is None:
        return [LuaBool(True)]

    # This function is equivalent to the ISO C function system.
    # It passes command to be executed by an operating system shell.
    if not isinstance(command, LuaString):
        raise LuaError("'command' must be a string")
    retcode = subprocess.call(
        command.content.decode("utf-8"),
        shell=True,
    )
    # Its first result is true if the command terminated successfully,
    # or fail otherwise.
    # After this first result the function returns a string plus a number,
    # as follows:
    #     "exit": the command terminated normally; the following number is
    #             the exit status of the command.
    #     "signal": the command was terminated by a signal; the following
    #               number is the signal that terminated the command.
    return [
        LuaBool(True) if retcode == 0 else FAIL,
        str_to_lua_string("exit" if retcode >= 0 else "signal"),
        LuaNumber(abs(retcode)),
    ]


@lua_function(name="exit")
def lf_os_exit(code=None, close=None, /) -> PyLuaRet:
    return os_exit(code, close)


def os_exit(code=None, close=None, /) -> PyLuaRet:
    # Calls the ISO C function exit to terminate the host program.
    # If code is true, the returned status is EXIT_SUCCESS;
    # if code is false, the returned status is EXIT_FAILURE;
    # if code is a number, the returned status is this number.
    # The default value for code is true.
    if code is None:
        code = 0
    elif isinstance(code, LuaNumber):
        code = code.value
    elif isinstance(code, LuaBool):
        code = 0 if code.true else 1
    else:
        raise LuaError("'code' must be a number or a boolean")

    # If the optional second argument close is true, the function closes the
    # Lua state before exiting (see lua_close).
    if close == LuaBool(True):
        sys.exit(code)
    else:
        os._exit(code)
    return []


@lua_function(name="getenv")
def lf_os_getenv(varname, /) -> PyLuaRet:
    return os_getenv(varname)


def os_getenv(varname, /) -> PyLuaRet:
    #  Returns the value of the process environment variable varname or fail
    #  if the variable is not defined.
    if not isinstance(varname, LuaString):
        raise LuaError("'varname' must be a string")
    value = os.getenv(varname.content.decode("utf-8"))
    if value is None:
        return [FAIL]
    return [str_to_lua_string(value)]


@lua_function(name="remove")
def lf_os_remove(filename, /) -> PyLuaRet:
    return os_remove(filename)


def os_remove(filename, /) -> PyLuaRet:
    # Deletes the file (or empty directory, on POSIX systems) with the
    # given name.
    if not isinstance(filename, LuaString):
        raise LuaError("'filename' must be a string")
    try:
        try:
            os.unlink(filename.content)
        except OSError as e:
            if os.name == "posix" and e.errno == 21:
                os.rmdir(filename.content)
            else:
                raise e
    except OSError as e:
        # If this function fails, it returns fail plus a string describing
        # the error and the error code.
        return _oserror_to_errtuple(e)
    # Otherwise, it returns true.
    return [LuaBool(True)]


@lua_function(name="rename")
def lf_os_rename(oldname, newname, /) -> PyLuaRet:
    return os_rename(oldname, newname)


def os_rename(oldname, newname, /) -> PyLuaRet:
    # Renames the file or directory named oldname to newname.
    if not isinstance(oldname, LuaString):
        raise LuaError("'oldname' must be a string")
    if not isinstance(newname, LuaString):
        raise LuaError("'newname' must be a string")
    try:
        os.rename(oldname.content, newname.content)
    except OSError as e:
        # If this function fails, it returns fail,
        # plus a string describing the error and the error code.
        return _oserror_to_errtuple(e)
    # Otherwise, it returns true.
    return [LuaBool(True)]


@lua_function(name="setlocale")
def lf_os_setlocale(locale, category=None, /) -> PyLuaRet:
    return os_setlocale(locale, category)


def os_setlocale(locale, category=None, /) -> PyLuaRet:
    # category is an optional string describing which category to change:
    # "all", "collate", "ctype", "monetary", "numeric", or "time";
    # the default category is "all".
    if category is None:
        category = lc.LC_ALL
    else:
        category = _get_category_from_luastr(category)
    # When called with nil as the first argument, this function only returns
    # the name of the current locale for the given category.
    if locale is LuaNil:
        current_lc = lc.getlocale(category)
        return [
            py2lua(current_lc[0]),
            py2lua(current_lc[1]),
        ]

    # Sets the current locale of the program.
    # locale is a system-dependent string specifying a locale;
    if not isinstance(locale, LuaString):
        raise LuaError("'locale' must be a string")
    # If locale is the empty string, the current locale is set to an
    # implementation-defined native locale.
    if not locale.content:
        locale = None
    else:
        locale = locale.content.decode("utf-8")
    # If locale is the string "C", the current locale is set to the standard
    # C locale.
    try:
        new_locale_name = lc.setlocale(category, locale)
        # The function returns the name of the new locale,
        # or fail if the request cannot be honored.
    except lc.Error:
        return [FAIL]
    else:
        return [str_to_lua_string(new_locale_name)]


@lua_function(name="time")
def lf_os_time(table=None, /) -> PyLuaRet:
    return os_time(table)


def os_time(table=None, /) -> PyLuaRet:
    raise NotImplementedError()


@lua_function(name="tmpname")
def lf_os_tmpname() -> PyLuaRet:
    return os_tmpname()


def os_tmpname() -> PyLuaRet:
    fd, name = tempfile.mkstemp()
    return [str_to_lua_string(name)]


class OSLibrary(LibraryProvider):
    def provide(self, global_table: LuaTable) -> None:
        os_table = LuaTable()
        global_table.rawput(LuaString(b"os"), os_table)

        for name_of_global, value_of_global in globals().items():
            if name_of_global.startswith("lf_os_"):
                assert isinstance(value_of_global, LuaFunction)
                assert value_of_global.name
                os_table.rawput(
                    LuaString(value_of_global.name.encode("ascii")),
                    value_of_global,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

import attrs

if TYPE_CHECKING:
    from mehtap.ast_nodes import Name
    from mehtap.values import LuaValue


class LuaError(BaseException):
    __slots__ = ["message", "level"]

    message: LuaValue
    level: int
    caused_by: Exception | None = None
    traceback_messages: list[str]

    def __init__(
        self,
        message: LuaValue | str,
        level: int = 1,
        caused_by: Exception | None = None
    ):
        if isinstance(message, str):
            from mehtap.values import LuaString
            message = LuaString(message.encode("ascii"))
        self.message = message
        self.level = level
        self.caused_by = caused_by
        self.traceback_messages = []
        super().__init__(message, level)

    def __repr__(self):
        return f"<error level {self.level}: {self.message}>"

    def __str__(self):
        return str(self.message)

    def push_tb(
        self,
        tb: str,
        *,
        file: str | None = None,
        line: int | None = None
    ):
        if file is not None and line is not None:
            self.traceback_messages.append(f'{file}:{line}: {tb}')
            return
        if file is not None and line is None:
            self.traceback_messages.append(f'{file}: {tb}')
            return
        if file is None and line is not None:
            self.traceback_messages.append(f'{line}: {tb}')
            return
        self.traceback_messages.append(tb)


@attrs.define(slots=True)
class FlowControlException(BaseException):
    pass


@attrs.define(slots=True)
class BreakException(FlowControlException):
    pass


@attrs.define(slots=True)
class ReturnException(FlowControlException):
    values: list[LuaValue] | None = None


@attrs.define(slots=True)
class GotoException(FlowControlException):
    label: Name
